import unittest
import logging
import sys
import os
import json
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint
import langchain
from langchain.agents import Tool, initialize_agent, load_tools
from langchain.llms import OpenAI
from typing import Any, List, Mapping, Optional
from langchain.llms.base import LLM

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader
from graphsignal.traces import DEFAULT_OPTIONS
from graphsignal.recorders.langchain_recorder import LangChainRecorder

logger = logging.getLogger('graphsignal')


class DummyLLM(LLM):
    @property
    def _llm_type(self) -> str:
        return "dummy"

    def _call(self, prompt: str, stop: Optional[List[str]] = None) -> str:
        return 'Final Answer:42'

    @property
    def _identifying_params(self) -> Mapping[str, Any]:
        return {}


class LangChainRecorderTest(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    async def asyncTearDown(self):
        graphsignal.shutdown()

    async def test_record(self):
        recorder = LangChainRecorder()
        recorder.setup()
        proto = signals_pb2.Trace()
        context = {}
        recorder.on_trace_start(proto, context, DEFAULT_OPTIONS)
        recorder.on_trace_stop(proto, context, DEFAULT_OPTIONS)
        recorder.on_trace_read(proto, context, DEFAULT_OPTIONS)

        self.assertEqual(proto.frameworks[0].name, 'LangChain')

    @patch.object(Uploader, 'upload_trace')
    async def test_chain(self, mocked_upload_trace):
        os.environ['OPENAI_API_KEY'] = 'dummy-api-key'
        llm = OpenAI(temperature=0)
        llm = DummyLLM()
        tools = load_tools(["llm-math"], llm=llm)
        agent = initialize_agent(
            tools, llm, agent="zero-shot-react-description", verbose=True
        )
        agent.run("What is 2 raised to .123243 power?")

        t3 = mocked_upload_trace.call_args_list[0][0][0]
        t2 = mocked_upload_trace.call_args_list[1][0][0]
        t1 = mocked_upload_trace.call_args_list[2][0][0]

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(proto))

        self.assertEqual(t1.frameworks[0].name, 'LangChain')
        self.assertEqual(t1.labels, ['root'])

        self.assertEqual(find_data_count(t1, 'inputs', 'byte_count'), 34.0)
        self.assertEqual(find_data_count(t1, 'inputs', 'element_count'), 1.0)
        self.assertEqual(find_data_count(t1, 'outputs', 'byte_count'), 2.0)
        self.assertEqual(find_data_count(t1, 'outputs', 'element_count'), 1.0)

        self.assertEqual(find_tag(t1, 'endpoint'), 'langchain.chains.AgentExecutor')

        self.assertEqual(find_tag(t2, 'endpoint'), 'langchain.chains.LLMChain')
        self.assertEqual(t2.span.parent_trace_id, t1.trace_id)
        self.assertEqual(t2.span.root_trace_id, t1.trace_id)

        self.assertEqual(find_tag(t3, 'endpoint'), 'langchain.llms.DummyLLM')
        self.assertEqual(t3.span.parent_trace_id, t2.trace_id)
        self.assertEqual(t3.span.root_trace_id, t1.trace_id)


def find_tag(proto, name):
    for tag in proto.tags:
        if tag.key == name:
            return tag.value
    return None


def find_data_count(proto, data_name, count_name):
    for data_stats in proto.data_profile:
        if data_stats.data_name == data_name:
            for data_count in data_stats.counts:
                if data_count.name == count_name:
                    return data_count.count
    return None



import unittest
import logging
import sys
import os
import json
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint

import graphsignal
from graphsignal.recorders.instrumentation import patch_method, instrument_method, read_args
from graphsignal.uploader import Uploader

logger = logging.getLogger('graphsignal')


class Dummy:
    def __init__(self):
        pass

    def test(self, a, b, c=None):
        return a + 1

    def test_exc(self):
        raise Exception('exc1')

    def test_gen(self):
        for i in range(2):
            yield 'item' + str(i)


class RecorderUtilsTest(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    async def asyncTearDown(self):
        graphsignal.shutdown()

    @patch.object(Uploader, 'upload_trace')
    async def test_instrument_method(self, mocked_upload_trace):
        obj = Dummy()

        trace_func_called = False
        def trace_func(trace, args, kwargs, ret, exc):
            nonlocal trace_func_called
            trace_func_called = True

        instrument_method(obj, 'test', 'ep1', trace_func)

        obj.test(1, 2, c=3)

        proto = mocked_upload_trace.call_args[0][0]

        self.assertTrue(trace_func_called)
        self.assertEqual(proto.tags[1].value, 'ep1')

    @patch.object(Uploader, 'upload_trace')
    async def test_instrument_method_generator(self, mocked_upload_trace):
        obj = Dummy()

        trace_func_called = None
        def trace_func(trace, args, kwargs, ret, exc):
            nonlocal trace_func_called
            trace_func_called = True

        instrument_method(obj, 'test_gen', 'ep1', trace_func)

        for item in obj.test_gen():
            pass

        proto = mocked_upload_trace.call_args[0][0]

        self.assertTrue(trace_func_called)
        self.assertEqual(proto.tags[1].value, 'ep1')

    async def test_patch_method(self):
        obj = Dummy()

        before_func_called = False
        def before_func(args, kwargs):
            nonlocal before_func_called
            before_func_called = True
            self.assertEqual(args, (1, 2))
            self.assertEqual(kwargs, {'c': 3})
            return dict(d=1)

        after_func_called = False
        def after_func(args, kwargs, ret, exc, context):
            nonlocal after_func_called
            after_func_called = True
            self.assertEqual(args, (1, 2))
            self.assertEqual(kwargs, {'c': 3})
            self.assertEqual(ret, 2)
            self.assertIsNone(exc)
            self.assertEqual(context['d'], 1)

        self.assertTrue(patch_method(obj, 'test', before_func=before_func, after_func=after_func))

        obj.test(1, 2, c=3)

        self.assertTrue(before_func_called)
        self.assertTrue(after_func_called)

    async def test_patch_method_exc(self):
        obj = Dummy()

        after_func_called = False
        def after_func(args, kwargs, ret, exc, context):
            nonlocal after_func_called
            after_func_called = True
            self.assertEqual(str(exc), 'exc1')

        self.assertTrue(patch_method(obj, 'test_exc', after_func=after_func))

        with self.assertRaises(Exception) as context:
            obj.test_exc()

        self.assertTrue(after_func_called)

    async def test_patch_method_generator(self):
        obj = Dummy()

        yield_func_called = False
        def yield_func(idx, item, context):
            nonlocal yield_func_called
            yield_func_called = True
            if idx != -1:
                self.assertTrue(item in ('item0', 'item1'))

        self.assertTrue(patch_method(obj, 'test_gen', yield_func=yield_func))

        for item in obj.test_gen():
            pass

        self.assertTrue(yield_func_called)

    async def test_read_args(self):
        def test(*args, **kwargs):
            values = read_args(args, kwargs, ['a', 'b', 'c'])
            self.assertEqual(values, {'a': 1, 'b': 2, 'c': 3})

        test(1, 2, c=3)


import logging
import sys
import importlib
from abc import ABC, abstractmethod
import functools

import graphsignal

logger = logging.getLogger('graphsignal')


class DataStats:
    __slots__ = ['type_name', 'shape', 'counts']

    def __init__(self, type_name=None, shape=None, counts=None):
        self.type_name = type_name
        self.shape = shape
        if counts is not None:
            self.counts = {name:value for name, value in counts.items() if value > 0}
        else:
            self.counts = {}

    def __repr__(self):
        return 'DataStats(type_name={0}, shape={1}, counts={2})'.format(
            self.type_name, self.shape, self.counts)


class DataSample:
    __slots__ = ['content_type', 'content_bytes']

    def __init__(self, content_type=None, content_bytes=None, is_partial=False):
        self.content_type = content_type
        self.content_bytes = content_bytes

    def __repr__(self):
        return 'DataSample(content_type={0}, content_bytes={1})'.format(
            self.content_type, self.content_bytes)


class BaseDataProfiler(ABC):
    def __init__(self):
        self._module = None
        self._is_checked = False

    def check_module(self, name):
        if not self._is_checked:
            self._is_checked = True
            if name in sys.modules:
                try:
                    self._module = importlib.import_module(name)
                except ImportError:
                    logger.error('Error importing {0}'.format(name))            
        return self._module

    @abstractmethod
    def is_instance(self, data):
        pass

    @abstractmethod
    def compute_stats(self, data):
        pass

    @abstractmethod
    def encode_sample(self, data):
        pass


import unittest
import logging
import sys
import os
import json
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint
import torch

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader
from graphsignal.traces import DEFAULT_OPTIONS
from graphsignal.recorders.pytorch_recorder import PyTorchRecorder

logger = logging.getLogger('graphsignal')


class PyTorchRecorderTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_record(self):
        recorder = PyTorchRecorder()
        recorder.setup()

        mem1 = torch.rand(10000)
        if torch.cuda.is_available():
            mem1 = mem1.cuda()
            torch.cuda.synchronize()

        proto = signals_pb2.Trace()
        context = {}
        recorder.on_trace_start(proto, context, DEFAULT_OPTIONS)

        mem2 = torch.rand(10000)
        if torch.cuda.is_available():
            mem2 = mem2.cuda()
        del mem2
        if torch.cuda.is_available():
            torch.cuda.synchronize()

        recorder.on_trace_stop(proto, context, DEFAULT_OPTIONS)
        recorder.on_trace_read(proto, context, DEFAULT_OPTIONS)

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(proto))

        self.assertEqual(proto.frameworks[0].name, 'PyTorch')

        self.assertEqual(proto.frameworks[0].params[0].name, 'torch.cuda.is_available')
        self.assertEqual(proto.frameworks[0].params[0].value, str(torch.cuda.is_available()))

        if torch.cuda.is_available():
            self.assertEqual(proto.alloc_summary[0].allocator_type, signals_pb2.MemoryAllocation.AllocatorType.PYTORCH_CUDA_ALLOCATOR)
            self.assertEqual(proto.alloc_summary[0].device_idx, 0)
            self.assertTrue(proto.alloc_summary[0].allocated_size > 0)
            self.assertTrue(proto.alloc_summary[0].freed_size > 0)
            self.assertEqual(proto.alloc_summary[0].num_allocations, 1)



import logging
import sys
import os
import time
import types
import openai

import graphsignal
from graphsignal.traces import TraceOptions
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.recorders.instrumentation import instrument_method, uninstrument_method, read_args
from graphsignal.proto_utils import parse_semver, compare_semver
from graphsignal.proto import signals_pb2
from graphsignal.proto_utils import add_framework_param, add_driver

logger = logging.getLogger('graphsignal')

class OpenAIRecorder(BaseRecorder):
    def __init__(self):
        self._framework = None

    def setup(self):
        if not graphsignal._agent.auto_instrument:
            return

        self._framework = signals_pb2.FrameworkInfo()
        self._framework.name = 'OpenAI Python Library'
        parse_semver(self._framework.version, openai.version.VERSION)

        if compare_semver(self._framework.version, (0, 26, 0)) < 1:
            logger.debug('OpenAI tracing is only supported for >= 0.26.0.')
            return

        self._api_base = openai.api_base

        instrument_method(openai.Completion, 'create', f'{self._api_base}/completions', self.trace_completion, self.trace_completion_data)
        instrument_method(openai.Completion, 'acreate', f'{self._api_base}/completions', self.trace_completion, self.trace_completion_data)
        instrument_method(openai.ChatCompletion, 'create', f'{self._api_base}/chat/completions', self.trace_chat_completion, self.trace_chat_completion_data)
        instrument_method(openai.ChatCompletion, 'acreate', f'{self._api_base}/chat/completions', self.trace_chat_completion, self.trace_chat_completion_data)
        instrument_method(openai.Edit, 'create', f'{self._api_base}/edits', self.trace_edits)
        instrument_method(openai.Edit, 'acreate', f'{self._api_base}/edits', self.trace_edits)
        instrument_method(openai.Embedding, 'create', f'{self._api_base}/embeddings', self.trace_embedding)
        instrument_method(openai.Embedding, 'acreate', f'{self._api_base}/embeddings', self.trace_embedding)
        instrument_method(openai.Image, 'create', f'{self._api_base}/images/generations', self.trace_image_generation)
        instrument_method(openai.Image, 'acreate', f'{self._api_base}/images/generations', self.trace_image_generation)
        instrument_method(openai.Image, 'create_variation', f'{self._api_base}/images/variations', self.trace_image_generation)
        instrument_method(openai.Image, 'acreate_variation', f'{self._api_base}/images/variations', self.trace_image_generation)
        instrument_method(openai.Image, 'create_edit', f'{self._api_base}/images/edits', self.trace_image_generation)
        instrument_method(openai.Image, 'acreate_edit', f'{self._api_base}/images/edits', self.trace_image_generation)
        instrument_method(openai.Audio, 'transcribe', f'{self._api_base}/audio/transcriptions', self.trace_audio_transcription)
        instrument_method(openai.Audio, 'atranscribe', f'{self._api_base}/audio/transcriptions', self.trace_audio_transcription)
        instrument_method(openai.Audio, 'translate', f'{self._api_base}/audio/translations', self.trace_audio_translation)
        instrument_method(openai.Audio, 'atranslate', f'{self._api_base}/audio/translations', self.trace_audio_translation)
        instrument_method(openai.Moderation, 'create', f'{self._api_base}/moderations', self.trace_moderation)
        instrument_method(openai.Moderation, 'acreate', f'{self._api_base}/moderations', self.trace_moderation)

    def shutdown(self):
        uninstrument_method(openai.Completion, 'create', f'{self._api_base}/completions')
        uninstrument_method(openai.Completion, 'acreate', f'{self._api_base}/completions')
        uninstrument_method(openai.ChatCompletion, 'create', f'{self._api_base}/chat/completions')
        uninstrument_method(openai.ChatCompletion, 'acreate', 'openai.ChatCompletion.acreate')
        uninstrument_method(openai.Edit, 'create', f'{self._api_base}/edits')
        uninstrument_method(openai.Edit, 'acreate', f'{self._api_base}/edits')
        uninstrument_method(openai.Embedding, 'create', f'{self._api_base}/embeddings')
        uninstrument_method(openai.Embedding, 'acreate', f'{self._api_base}/embeddings')
        uninstrument_method(openai.Image, 'create', f'{self._api_base}/images/generations')
        uninstrument_method(openai.Image, 'acreate', f'{self._api_base}/images/generations')
        uninstrument_method(openai.Image, 'create_variation', f'{self._api_base}/images/variations')
        uninstrument_method(openai.Image, 'acreate_variation', f'{self._api_base}/images/variations')
        uninstrument_method(openai.Image, 'create_edit', f'{self._api_base}/images/edits')
        uninstrument_method(openai.Image, 'acreate_edit', f'{self._api_base}/images/edits')
        uninstrument_method(openai.Audio, 'transcribe', f'{self._api_base}/audio/transcriptions')
        uninstrument_method(openai.Audio, 'atranscribe', f'{self._api_base}/audio/transcriptions')
        uninstrument_method(openai.Audio, 'translate', f'{self._api_base}/audio/translations')
        uninstrument_method(openai.Audio, 'atranslate', f'{self._api_base}/audio/translations')
        uninstrument_method(openai.Moderation, 'create', f'{self._api_base}/moderations')
        uninstrument_method(openai.Moderation, 'acreate', f'{self._api_base}/moderations')

    def trace_completion(self, trace, args, kwargs, ret, exc):
        params = kwargs # no positional args

        if 'model' in params:
            trace.set_tag('model', params['model'])

        param_names = [
            'model',
            'max_tokens',
            'temperature',
            'top_p',
            'n',
            'stream',
            'logprobs',
            'echo',
            'stop',
            'presence_penalty',
            'frequency_penalty',
            'best_of'
        ]
        for param_name in param_names:
            if param_name in params:
                trace.set_param(param_name, params[param_name])

        if 'stream' in params and params['stream']:
            if 'prompt' in params:
                trace.set_data('prompt', params['prompt'])
            return

        if ret and 'model' in ret:
            trace.set_tag('model', ret['model'])

        prompt_usage = {}
        completion_usage = {
            'finish_reason_stop': 0,
            'finish_reason_length': 0
        }
        if ret and 'usage' in ret:
            if 'prompt_tokens' in ret['usage']:
                prompt_usage['token_count'] = ret['usage']['prompt_tokens']
            if 'completion_tokens' in ret['usage']:
                completion_usage['token_count'] = ret['usage']['completion_tokens']

        if 'prompt' in params:
            trace.set_data('prompt', params['prompt'], counts=prompt_usage)

        if ret and 'choices' in ret:
            completion = []
            for choice in ret['choices']:
                if 'finish_reason' in choice:
                    if choice['finish_reason'] == 'stop':
                        completion_usage['finish_reason_stop'] += 1
                    elif choice['finish_reason'] == 'length':
                        completion_usage['finish_reason_length'] += 1
                if 'text' in choice:
                    completion.append(choice['text'])
            trace.set_data('completion', completion, counts=completion_usage)

    def trace_completion_data(self, trace, item):
        completion_usage = {
            'finish_reason_stop': 0,
            'finish_reason_length': 0
        }
        if item and 'choices' in item:
            completion = []
            for choice in item['choices']:
                if 'finish_reason' in choice:
                    if choice['finish_reason'] == 'stop':
                        completion_usage['finish_reason_stop'] += 1
                    elif choice['finish_reason'] == 'length':
                        completion_usage['finish_reason_length'] += 1
                if 'text' in choice:
                    completion.append(choice['text'])

            trace.append_data('completion', completion, counts=completion_usage)

    def trace_chat_completion(self, trace, args, kwargs, ret, exc):
        params = kwargs # no positional args

        if 'model' in params:
            trace.set_tag('model', params['model'])

        param_names = [
            'model',
            'max_tokens',
            'temperature',
            'top_p',
            'n',
            'stream',
            'logprobs',
            'stop',
            'presence_penalty',
            'frequency_penalty',
            'best_of'
        ]
        for param_name in param_names:
            if param_name in params:
                trace.set_param(param_name, params[param_name])

        if 'stream' in params and params['stream']:
            if 'messages' in params:
                trace.set_data('messages', params['messages'])
            return

        if ret and 'model' in ret:
            trace.set_tag('model', ret['model'])

        prompt_usage = {}
        completion_usage = {
            'finish_reason_stop': 0,
            'finish_reason_length': 0
        }
        if ret and 'usage' in ret:
            if 'prompt_tokens' in ret['usage']:
                prompt_usage['token_count'] = ret['usage']['prompt_tokens']
            if 'completion_tokens' in ret['usage']:
                completion_usage['token_count'] = ret['usage']['completion_tokens']

        if 'messages' in params:
            trace.set_data('messages', params['messages'], counts=prompt_usage)

        if ret and 'choices' in ret:
            completion = []
            for choice in ret['choices']:
                if 'finish_reason' in choice:
                    if choice['finish_reason'] == 'stop':
                        completion_usage['finish_reason_stop'] += 1
                    elif choice['finish_reason'] == 'length':
                        completion_usage['finish_reason_length'] += 1
                if 'message' in choice and 'content' in choice['message']:
                    completion.append(choice['message']['content'])
            trace.set_data('completion', completion, counts=completion_usage)

    def trace_chat_completion_data(self, trace, item):
        completion_usage = {
            'finish_reason_stop': 0,
            'finish_reason_length': 0
        }
        if item and 'choices' in item:
            completion = []
            for choice in item['choices']:
                if 'finish_reason' in choice:
                    if choice['finish_reason'] == 'stop':
                        completion_usage['finish_reason_stop'] += 1
                    elif choice['finish_reason'] == 'length':
                        completion_usage['finish_reason_length'] += 1
                if 'delta' in choice and 'content' in choice['delta']:
                    completion.append(choice['delta']['content'])

            trace.append_data('completion', completion, counts=completion_usage)

    def trace_edits(self, trace, args, kwargs, ret, exc):
        params = kwargs # no positional args

        if 'model' in params:
            trace.set_tag('model', params['model'])

        param_names = [
            'model',
            'temperature',
            'top_p',
            'n'
        ]
        for param_name in param_names:
            if param_name in params:
                trace.set_param(param_name, params[param_name])

        prompt_usage = {}
        completion_usage = {}
        if 'usage' in ret:
            if 'prompt_tokens' in ret['usage']:
                prompt_usage['token_count'] = ret['usage']['prompt_tokens']
            if 'completion_tokens' in ret['usage']:
                completion_usage['token_count'] = ret['usage']['completion_tokens']

        if 'input' in params:
            trace.set_data('input', params['input'])

        if 'instruction' in params:
            trace.set_data('instruction', params['instruction'], counts=prompt_usage)

        if ret and 'choices' in ret:
            edits = []
            for choice in ret['choices']:
                if 'text' in choice:
                    edits.append(choice['text'])
            trace.set_data('edits', edits, counts=completion_usage)

    def trace_embedding(self, trace, args, kwargs, ret, exc):
        params = kwargs # no positional args

        if 'engine' in params:
            trace.set_tag('model', params['engine'])

        param_names = [
            'engine'
        ]
        for param_name in param_names:
            if param_name in params:
                trace.set_param(param_name, params[param_name])

        prompt_usage = {}
        if 'usage' in ret:
            if 'prompt_tokens' in ret['usage']:
                prompt_usage['token_count'] = ret['usage']['prompt_tokens']

        if 'input' in params:
            trace.set_data('input', params['input'], counts=prompt_usage)

        if ret and 'data' in ret:
            embedding = []
            for choice in ret['data']:
                if 'embedding' in choice:
                    embedding.append(choice['embedding'])
            trace.set_data('embedding', embedding)

    def trace_image_generation(self, trace, args, kwargs, ret, exc):
        params = kwargs # no positional args

        param_names = [
            'n',
            'size',
            'response_format'
        ]
        for param_name in param_names:
            if param_name in params:
                trace.set_param(param_name, params[param_name])

        if 'prompt' in params:
            trace.set_data('prompt', params['prompt'])

        if ret and 'data' in ret:
            image_data = []
            for image in ret['data']:
                if 'url' in image:
                    image_data.append(image['url'])
                elif 'b64_json' in image:
                    image_data.append(image['b64_json'])
            trace.set_data('image', image_data)

    def trace_audio_transcription(self, trace, args, kwargs, ret, exc):
        params = read_args(args, kwargs, ['model', 'file', 'prompt', 'response_format', 'temperature', 'language'])

        if 'model' in params:
            trace.set_tag('model', params['model'])

        param_names = [
            'model',
            'temperature',
            'response_format',
            'language'
        ]
        for param_name in param_names:
            if param_name in params:
                trace.set_param(param_name, params[param_name])

        if 'file' in params and hasattr(params['file'], 'name'):
            try:
                file_size = os.path.getsize(params['file'].name)
                trace.set_data('file', params['file'], counts={'byte_count': file_size})
            except OSError:
                pass

        if 'prompt' in params:
            trace.set_data('prompt', params['prompt'])

        if ret and 'text' in ret:
            trace.set_data('text', ret['text'])

    def trace_audio_translation(self, trace, args, kwargs, ret, exc):
        params = read_args(args, kwargs, ['model', 'file', 'prompt', 'response_format', 'temperature'])

        if 'model' in params:
            trace.set_tag('model', params['model'])

        param_names = [
            'model',
            'temperature',
            'response_format'
        ]
        for param_name in param_names:
            if param_name in params:
                trace.set_param(param_name, params[param_name])

        if 'file' in params and hasattr(params['file'], 'name'):
            try:
                file_size = os.path.getsize(params['file'].name)
                trace.set_data('file', params['file'], counts={'byte_count': file_size})
            except OSError:
                pass

        if 'prompt' in params:
            trace.set_data('prompt', params['prompt'])

        if ret and 'text' in ret:
            trace.set_data('text', ret['text'])

    def trace_moderation(self, trace, args, kwargs, ret, exc):
        params = read_args(args, kwargs, ['input', 'model'])

        if 'model' in params:
            trace.set_tag('model', params['model'])

        for param_name in params:
            if param_name in params:
                trace.set_param(param_name, params[param_name])

        if 'input' in params:
            trace.set_data('input', params['input'])

    def on_trace_start(self, proto, context, options):
        pass

    def on_trace_stop(self, proto, context, options):
        pass

    def on_trace_read(self, proto, context, options):
        if self._framework:
            proto.frameworks.append(self._framework)

    def on_metric_update(self):
        pass




import logging
import os
import sys
import platform
import time
import re
import multiprocessing
import socket
try:
    import resource
except ImportError:
    pass

import graphsignal
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.proto import signals_pb2
from graphsignal.proto_utils import parse_semver

logger = logging.getLogger('graphsignal')

OS_LINUX = (sys.platform.startswith('linux'))
OS_DARWIN = (sys.platform == 'darwin')
OS_WIN = (sys.platform == 'win32')
CPU_NAME_REGEXP = re.compile(r'Model name:\s+(.+)$', flags=re.MULTILINE)
CPU_NAME_MAC_REGEXP = re.compile(r'machdep\.cpu\.brand_string:\s+(.+)$', flags=re.MULTILINE)
VM_RSS_REGEXP = re.compile(r'VmRSS:\s+(\d+)\s+kB')
VM_SIZE_REGEXP = re.compile(r'VmSize:\s+(\d+)\s+kB')
MEM_TOTAL_REGEXP = re.compile(r'MemTotal:\s+(\d+)\s+kB')
MEM_FREE_REGEXP = re.compile(r'MemFree:\s+(\d+)\s+kB')

class ProcessRecorder(BaseRecorder):
    MIN_CPU_READ_INTERVAL_US = 1 * 1e6

    def __init__(self):
        self._last_read_sec = None
        self._last_cpu_time_us = None

    def on_trace_start(self, proto, context, options):
        pass

    def on_trace_stop(self, proto, context, options):
        pass

    def on_trace_read(self, proto, context, options):
        if not OS_WIN:
            rusage_self = resource.getrusage(resource.RUSAGE_SELF)

        if OS_LINUX:
            current_rss = _read_current_rss()
            vm_size = _read_vm_size()

        now = time.time()
        pid = os.getpid()

        node_usage = proto.node_usage
        process_usage = proto.process_usage

        process_usage.pid = pid

        if not OS_WIN:
            cpu_time_us = _rusage_cpu_time(rusage_self)
            if cpu_time_us is not None:
                if self._last_cpu_time_us is not None:
                    interval_us = (now - self._last_read_sec) * 1e6
                    if interval_us > ProcessRecorder.MIN_CPU_READ_INTERVAL_US:
                        cpu_diff_us = cpu_time_us - self._last_cpu_time_us
                        cpu_usage = (cpu_diff_us / interval_us) * 100
                        try:
                            cpu_usage = cpu_usage / multiprocessing.cpu_count()
                        except Exception:
                            pass
                        process_usage.cpu_usage_percent = cpu_usage

                if (self._last_read_sec is None or 
                        now - self._last_read_sec > ProcessRecorder.MIN_CPU_READ_INTERVAL_US):
                    self._last_read_sec = now
                    self._last_cpu_time_us = cpu_time_us

            if OS_DARWIN:
                max_rss = rusage_self.ru_maxrss
            else:
                max_rss = rusage_self.ru_maxrss * 1e3
            if max_rss is not None:
                process_usage.max_rss = int(max_rss)

        if OS_LINUX:
            if current_rss is not None:
                process_usage.current_rss = current_rss

            if vm_size is not None:
                process_usage.vm_size = vm_size

            mem_total = _read_mem_total()
            if mem_total is not None:
                node_usage.mem_total = mem_total
                mem_free = _read_mem_free()
                if mem_free is not None:
                    node_usage.mem_used = mem_total - mem_free

        try:
            node_usage.hostname = socket.gethostname()
            if node_usage.hostname:
                node_usage.ip_address = socket.gethostbyname(node_usage.hostname)
        except BaseException:
            logger.debug('Error reading hostname', exc_info=True)

        try:
            node_usage.platform = sys.platform
            node_usage.machine = platform.machine()
            if not OS_WIN:
                node_usage.os_name = os.uname().sysname
                node_usage.os_version = os.uname().release
        except BaseException:
            logger.error('Error reading node information', exc_info=True)

        try:
            process_usage.runtime = signals_pb2.ProcessUsage.Runtime.PYTHON
            process_usage.runtime_version.major = sys.version_info.major
            process_usage.runtime_version.minor = sys.version_info.minor
            process_usage.runtime_version.patch = sys.version_info.micro
            process_usage.runtime_impl = platform.python_implementation()
        except BaseException:
            logger.error('Error reading process information', exc_info=True)

    def on_metric_update(self):
        now = int(time.time())
        proto = signals_pb2.Trace()
        self.on_trace_read(proto, {}, {})

        store = graphsignal._agent.metric_store()
        metric_tags = {'deployment': graphsignal._agent.deployment}
        if graphsignal._agent.hostname:
            metric_tags['hostname'] = graphsignal._agent.hostname

        if proto.process_usage.cpu_usage_percent > 0:
            store.set_gauge(
                scope='system', name='process_cpu_usage', tags=metric_tags, 
                value=proto.process_usage.cpu_usage_percent, update_ts=now, unit='%')
        if proto.process_usage.current_rss > 0:
            store.set_gauge(
                scope='system', name='process_memory', tags=metric_tags, 
                value=proto.process_usage.current_rss, update_ts=now, is_size=True)
        if proto.process_usage.vm_size > 0:
            store.set_gauge(
                scope='system', name='virtual_memory', tags=metric_tags, 
                value=proto.process_usage.vm_size, update_ts=now, is_size=True)
        if proto.node_usage.mem_used > 0:
            store.set_gauge(
                scope='system', name='node_memory_used', tags=metric_tags, 
                value=proto.node_usage.mem_used, update_ts=now, is_size=True)


def _rusage_cpu_time(rusage):
    return int((rusage.ru_utime + rusage.ru_stime) * 1e6)  # microseconds


def _read_current_rss():
    pid = os.getpid()

    try:
        f = open('/proc/{0}/status'.format(os.getpid()))
        output = f.read()
        f.close()

        match = VM_RSS_REGEXP.search(output)
        if match:
            return int(float(match.group(1)) * 1e3)
    except Exception:
        pass

    return None


def _read_vm_size():
    pid = os.getpid()

    try:
        f = open('/proc/{0}/status'.format(os.getpid()))
        output = f.read()
        f.close()

        match = VM_SIZE_REGEXP.search(output)
        if match:
            return int(float(match.group(1)) * 1e3)
    except BaseException:
        pass

    return None


def _read_mem_total():
    try:
        f = open('/proc/meminfo')
        output = f.read()
        f.close()

        match = MEM_TOTAL_REGEXP.search(output)
        if match:
            return int(float(match.group(1)) * 1e3)
    except BaseException:
        pass

    return None


def _read_mem_free():
    try:
        f = open('/proc/meminfo')
        output = f.read()
        f.close()

        match = MEM_FREE_REGEXP.search(output)
        if match:
            return int(float(match.group(1)) * 1e3)
    except BaseException:
        pass

    return None



import logging
import os
import sys
import yappi

import graphsignal
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.proto_utils import parse_semver
from graphsignal.proto import signals_pb2
from graphsignal.proto_utils import add_framework_param

logger = logging.getLogger('graphsignal')

class YappiRecorder(BaseRecorder):
    def __init__(self):
        self._exclude_path = os.path.dirname(os.path.realpath(graphsignal.__file__))

    def setup(self):
        pass

    def on_trace_start(self, proto, context, options):
        if not options.enable_profiling:
            return

        if 'torch' in sys.modules:
            return

        yappi.set_clock_type("wall")
        yappi.start()
        context['is_yappi_profiling'] = True

    def on_trace_stop(self, proto, context, options):
        if not context.get('is_yappi_profiling', False):
            return

        if yappi.is_running():
            yappi.stop()

    def on_trace_read(self, proto, context, options):
        if not context.get('is_yappi_profiling', False):
            return

        self._convert_to_profile(proto, yappi.get_func_stats())
        yappi.clear_stats()

    def on_metric_update(self):
        pass

    def _convert_to_profile(self, proto, ystats):
        stats = yappi.convert2pstats(ystats)

        stats.sort_stats('cumtime')
        func_keys = stats.fcn_list[:] if stats.fcn_list else list(stats.stats.keys())

        func_list = []
        total_tt = 0
        visited = {}
        for func_key in func_keys:
            func_stats = stats.stats[func_key]
            file_name, line_num, func_name = func_key
            if self._has_exclude_func(stats.stats, func_key, visited):
                continue
            cc, nc, tt, ct, _ = func_stats
            func_list.append((file_name, line_num, func_name, nc, tt, ct))
            total_tt += tt

        for file_name, line_num, func_name, nc, tt, ct in func_list[:250]:
            op_stats = proto.op_profile.add()
            op_stats.op_type = signals_pb2.OpStats.OpType.PYTHON_OP
            op_stats.op_name = _format_frame(file_name, line_num, func_name)
            op_stats.count = int(nc)
            op_stats.host_time_ns = _ns(ct)
            op_stats.self_host_time_ns = _ns(tt)
            if total_tt > 0:
                op_stats.self_host_time_percent = tt / total_tt * 100

        proto.labels.append('profiled')

    def _has_exclude_func(self, stats, func_key, visited):
        if func_key in visited:
            return visited[func_key]

        file_name, _, _ = func_key
        _, _, _, _, callers = stats[func_key]

        if file_name.startswith(self._exclude_path):
            visited[func_key] = True
            return True
        else:
            visited[func_key] = False

        for caller_key in callers.keys():
            if self._has_exclude_func(stats, caller_key, visited):
                visited[func_key] = True
                return True

        return False


def _ns(sec):
    return int(sec * 1e9)


def _format_frame(file_name, line_num, func_name):
    if file_name == '~':
        file_name = ''

    if file_name and line_num and func_name:
        return '{func_name} ({file_name}:{line_num})'.format(
            file_name=file_name,
            func_name=func_name,
            line_num=line_num)
    elif file_name and func_name:
        return '{func_name} ({file_name})'.format(
            file_name=file_name,
            func_name=func_name)
    elif func_name:
        return func_name
    else:
        return 'unknown'


import logging
import onnxruntime

import graphsignal
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.proto_utils import parse_semver
from graphsignal.proto import signals_pb2

logger = logging.getLogger('graphsignal')

class ONNXRuntimeRecorder(BaseRecorder):
    def __init__(self):
        self._framework = None

    def setup(self):
        self._framework = signals_pb2.FrameworkInfo()
        self._framework.name = 'ONNX Runtime'
        parse_semver(self._framework.version, onnxruntime.__version__)

    def on_trace_start(self, proto, context, options):
        pass

    def on_trace_stop(self, proto, context, options):
        pass

    def on_trace_read(self, proto, context, options):
        if self._framework:
            proto.frameworks.append(self._framework)

    def on_metric_update(self):
        pass



import unittest
import logging
import sys
from unittest.mock import patch, Mock
import torch
from google.protobuf.json_format import MessageToJson
import pprint
import time
import socket

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.traces import DEFAULT_OPTIONS
from graphsignal.recorders.nvml_recorder import NVMLRecorder

logger = logging.getLogger('graphsignal')


class NVMLRecorderTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_record(self):
        recorder = NVMLRecorder()
        recorder.setup()

        proto = signals_pb2.Trace()
        context = {}
        recorder.on_trace_start(proto, context, DEFAULT_OPTIONS)
        recorder.on_trace_stop(proto, context, DEFAULT_OPTIONS)
        recorder.on_trace_read(proto, context, DEFAULT_OPTIONS)

        model = torch.nn.Linear(1, 1)
        if torch.cuda.is_available():
            model = model.cuda()

        proto = signals_pb2.Trace()
        context = {}
        recorder.on_trace_start(proto, context, DEFAULT_OPTIONS)

        x = torch.arange(-50, 50, 0.00001).view(-1, 1)
        if torch.cuda.is_available():
            x = x.cuda()
        pred = model(x)

        proto = signals_pb2.Trace()
        recorder.on_trace_stop(proto, context, DEFAULT_OPTIONS)
        recorder.on_trace_read(proto, context, DEFAULT_OPTIONS)

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(proto))

        if len(proto.device_usage) > 0:
            self.assertTrue(proto.node_usage.num_devices > 0)

            self.assertEqual(proto.node_usage.drivers[0].name, 'CUDA')
            self.assertIsNotNone(proto.node_usage.drivers[0].version)

            device_usage = proto.device_usage[0]
            self.assertEqual(device_usage.device_type, signals_pb2.DeviceUsage.DeviceType.GPU_DEVICE)
            self.assertNotEqual(device_usage.device_id, 0)
            self.assertNotEqual(device_usage.device_id, '')
            self.assertNotEqual(device_usage.device_name, '')
            self.assertNotEqual(device_usage.architecture, '')
            self.assertTrue(device_usage.compute_capability.major > 0)
            self.assertTrue(device_usage.mem_total > 0)
            self.assertTrue(device_usage.mem_used > 0)
            self.assertTrue(device_usage.mem_free > 0)
            self.assertTrue(device_usage.gpu_utilization_percent > 0)
            #self.assertTrue(device_usage.mem_access_percent > 0)
            self.assertTrue(device_usage.gpu_temp_c > 0)
            self.assertTrue(device_usage.power_usage_w > 0)
            #self.assertTrue(device_usage.fan_speed_percent > 0)
            self.assertTrue(device_usage.gpu_temp_c > 0)
            self.assertTrue(device_usage.power_usage_w > 0)

            recorder.on_metric_update()

            store = graphsignal._agent.metric_store()
            metric_tags =  {'deployment': 'd1', 'hostname': socket.gethostname(), 'device': 0}
            key = store.metric_key('system', 'gpu_utilization', metric_tags)
            self.assertTrue(store.metrics[key].gauge > 0)
            key = store.metric_key('system', 'mxu_utilization', metric_tags)
            if key in store.metrics:
                self.assertTrue(store.metrics[key].gauge > 0)
            key = store.metric_key('system', 'device_memory_access', metric_tags)
            if key in store.metrics:
                self.assertTrue(store.metrics[key].gauge > 0)
            key = store.metric_key('system', 'device_memory_used', metric_tags)
            self.assertTrue(store.metrics[key].gauge > 0)
            key = store.metric_key('system', 'nvlink_throughput_data_tx_kibs', metric_tags)
            if key in store.metrics:
                self.assertTrue(store.metrics[key].gauge > 0)
            key = store.metric_key('system', 'nvlink_throughput_data_rx_kibs', metric_tags)
            if key in store.metrics:
                self.assertTrue(store.metrics[key].gauge > 0)
            key = store.metric_key('system', 'gpu_temp_c', metric_tags)
            self.assertTrue(store.metrics[key].gauge > 0)
            key = store.metric_key('system', 'power_usage_w', metric_tags)
            self.assertTrue(store.metrics[key].gauge > 0)



import logging
import functools

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.data.base_data_profiler import BaseDataProfiler, DataStats

logger = logging.getLogger('graphsignal')


class TFTensorProfiler(BaseDataProfiler):
    def __init__(self):
        super().__init__()

    def is_instance(self, data):
        tf = self.check_module('tensorflow')
        return tf is not None and tf.is_tensor(data)

    def compute_stats(self, data):
        tf = self.check_module('tensorflow')
        counts = {}
        counts['element_count'] = functools.reduce(lambda x, y: x*y, list(data.get_shape()))
        counts['byte_count'] = counts['element_count'] * data.dtype.size
        counts['nan_count'] = int(tf.math.count_nonzero(tf.math.is_nan(data)))
        counts['inf_count'] = int(tf.math.count_nonzero(tf.math.is_inf(data)))
        counts['zero_count'] = int(tf.math.count_nonzero(tf.math.equal(data, 0)))
        counts['negative_count'] = int(tf.math.count_nonzero(tf.math.less(data, 0)))
        counts['positive_count'] = int(tf.math.count_nonzero(tf.math.greater(data, 0)))
        return DataStats(type_name='tf.Tensor', shape=list(data.get_shape()), counts=counts)

    def encode_sample(self, data):
        return None



import logging
import sys
import time
import uuid
import hashlib
import importlib
import socket
import contextvars
import importlib.abc

from graphsignal import version
from graphsignal.uploader import Uploader
from graphsignal.metrics import MetricStore
from graphsignal.trace_sampler import TraceSampler
from graphsignal.proto import signals_pb2
from graphsignal.detectors.latency_outlier_detector import LatencyOutlierDetector
from graphsignal.detectors.missing_value_detector import MissingValueDetector
from graphsignal.proto_utils import parse_semver

logger = logging.getLogger('graphsignal')


RECORDER_SPECS = {
    '(default)': [
        ('graphsignal.recorders.cprofile_recorder', 'CProfileRecorder'),
        ('graphsignal.recorders.process_recorder', 'ProcessRecorder'),
        ('graphsignal.recorders.nvml_recorder', 'NVMLRecorder')],
    'torch': [
        ('graphsignal.recorders.pytorch_recorder', 'PyTorchRecorder'),
        ('graphsignal.recorders.kineto_recorder', 'KinetoRecorder')],
    'yappi': [('graphsignal.recorders.yappi_recorder', 'YappiRecorder')],
    'tensorflow': [('graphsignal.recorders.tensorflow_recorder', 'TensorFlowRecorder')],
    'jax': [('graphsignal.recorders.jax_recorder', 'JAXRecorder')],
    'onnxruntime': [('graphsignal.recorders.onnxruntime_recorder', 'ONNXRuntimeRecorder')],
    'deepspeed': [('graphsignal.recorders.deepspeed_recorder', 'DeepSpeedRecorder')],
    'openai': [('graphsignal.recorders.openai_recorder', 'OpenAIRecorder')],
    'langchain': [('graphsignal.recorders.langchain_recorder', 'LangChainRecorder')],
    'banana_dev': [('graphsignal.recorders.banana_recorder', 'BananaRecorder')]
}

class SourceLoaderWrapper(importlib.abc.SourceLoader):
    def __init__(self, loader, agent):
        self._loader = loader
        self._agent = agent

    def create_module(self, spec):
        return self._loader.create_module(spec)

    def exec_module(self, module):
        self._loader.exec_module(module)
        self._agent.initialize_recorders_for_module(module.__name__)

    def load_module(self, fullname):
        self._loader.load_module(fullname)

    def get_data(self, path):
        return self._loader.get_data(path)

    def get_filename(self, fullname):
        return self._loader.get_filename(fullname)


class SupportedModuleFinder(importlib.abc.MetaPathFinder):
    def __init__(self, agent):
        self._agent = agent
        self._disabled = False

    def find_spec(self, fullname, path=None, target=None):
        if self._disabled:
            return None

        if fullname in RECORDER_SPECS:
            try:
                self._disabled = True
                loader = importlib.util.find_spec(fullname).loader
                if loader is not None and isinstance(loader, importlib.abc.SourceLoader):
                    return importlib.util.spec_from_loader(fullname, SourceLoaderWrapper(loader, self._agent))
            except Exception:
                logger.error('Error patching spec for module %s', fullname, exc_info=True)
            finally:
                self._disabled = False

        return None


class Agent:
    METRIC_READ_INTERVAL_SEC = 10
    METRIC_UPLOAD_INTERVAL_SEC = 20

    def __init__(
            self, 
            api_key=None, 
            api_url=None, 
            deployment=None, 
            tags=None, 
            auto_instrument=True, 
            record_data_samples=False, 
            upload_on_shutdown=True, 
            debug_mode=False):
        self.api_key = api_key
        if api_url:
            self.api_url = api_url
        else:
            self.api_url = 'https://signal-api.graphsignal.com'
        self.deployment = deployment
        self.tags = tags
        self.context_tags = None
        self.auto_instrument = auto_instrument
        self.record_data_samples = record_data_samples
        self.upload_on_shutdown = upload_on_shutdown
        self.debug_mode = debug_mode
        self.hostname = None
        try:
            self.hostname = socket.gethostname()
        except BaseException:
            logger.debug('Error reading hostname', exc_info=True)

        self._uploader = None
        self._trace_samplers = None
        self._metric_store = None
        self._recorders = None
        self._lo_detectors = None
        self._mv_detector = None

        self._process_start_ms = int(time.time() * 1e3)

        self.last_metric_read_ts = 0
        self.last_metric_upload_ts = int(self._process_start_ms / 1e3)

    def setup(self):
        self._uploader = Uploader()
        self._uploader.setup()
        self._trace_samplers = {}
        self._metric_store = MetricStore()
        self._recorders = {}
        self._lo_detectors = {}

        # initialize context tags variable
        self.context_tags = contextvars.ContextVar('graphsignal_context_tags', default={})

        # initialize module wrappers
        self._module_finder = SupportedModuleFinder(self)
        sys.meta_path.insert(0, self._module_finder)

        # initialize default recorders and recorders for already loaded modules
        for module_name in RECORDER_SPECS.keys():
            if module_name == '(default)' or module_name in sys.modules:
                self.initialize_recorders_for_module(module_name)

    def shutdown(self):
        # Create snapshot signals to send final metrics.
        if self.upload_on_shutdown:
            if self._metric_store.has_unexported():
                metrics = self._metric_store.export()
                for metric in metrics:
                    self._uploader.upload_metric(metric)

        for recorder in self.recorders():
            recorder.shutdown()

        self.upload(block=True)

        self._recorders = None
        self._trace_samplers = None
        self._metric_store = None
        self._lo_detectors = None
        self._mv_detector = None
        self._uploader = None

        self.context_tags.set({})
        self.context_tags = None

        # remove module wrappers
        if self._module_finder in sys.meta_path:
            sys.meta_path.remove(self._module_finder)
        self._module_finder = None


    def uploader(self):
        return self._uploader

    def initialize_recorders_for_module(self, module_name):
        # check already loaded
        if module_name in self._recorders:
            return

        # check if supported
        if module_name not in RECORDER_SPECS:
            return

        # load recorder
        logger.debug('Initializing recorder for module: %s', module_name)
        specs = RECORDER_SPECS[module_name]
        self._recorders[module_name] = []
        for spec in specs:
            try:
                recorder_module = importlib.import_module(spec[0])
                recorder_class = getattr(recorder_module, spec[1])
                recorder = recorder_class()
                recorder.setup()
                self._recorders[module_name].append(recorder)
            except Exception:
                logger.error('Failed to initialize recorder for module: %s', module_name, exc_info=True)

    def recorders(self):
        for recorder_list in self._recorders.values():
            for recorder in recorder_list:
                yield recorder

    def trace_sampler(self, endpoint):
        if endpoint in self._trace_samplers:
            return self._trace_samplers[endpoint]
        else:
            trace_sampler = self._trace_samplers[endpoint] = TraceSampler()
            return trace_sampler

    def metric_store(self):
        return self._metric_store

    def lo_detector(self, endpoint):
        if endpoint in self._lo_detectors:
            return self._lo_detectors[endpoint]
        else:
            lo_detector = self._lo_detectors[endpoint] = LatencyOutlierDetector()
            return lo_detector

    def mv_detector(self):
        if self._mv_detector is None:
            self._mv_detector = MissingValueDetector()
        return self._mv_detector

    def emit_trace_start(self, proto, context, options):
        last_exc = None
        for recorder in self.recorders():
            try:
                recorder.on_trace_start(proto, context, options)
            except Exception as exc:
                last_exc = exc
        if last_exc:
            raise last_exc

    def emit_trace_stop(self, proto, context, options):
        last_exc = None
        for recorder in self.recorders():
            try:
                recorder.on_trace_stop(proto, context, options)
            except Exception as exc:
                last_exc = exc
        if last_exc:
            raise last_exc

    def emit_trace_read(self, proto, context, options):
        last_exc = None
        for recorder in self.recorders():
            try:
                recorder.on_trace_read(proto, context, options)
            except Exception as exc:
                last_exc = exc
        if last_exc:
            raise last_exc

    def emit_metric_update(self):
        last_exc = None
        for recorder in self.recorders():
            try:
                recorder.on_metric_update()
            except Exception as exc:
                last_exc = exc
        if last_exc:
            raise last_exc

    def create_trace_proto(self):
        proto = signals_pb2.Trace()
        proto.trace_id = _uuid_sha1(size=12)
        proto.tracer_info.tracer_type = signals_pb2.TracerInfo.TracerType.PYTHON_TRACER
        parse_semver(proto.tracer_info.version, version.__version__)
        return proto

    def upload(self, block=False):
        if block:
            self._uploader.flush()
        else:
            self._uploader.flush_in_thread()

    def check_metric_read_interval(self, now=None):
        if now is None:
            now = time.time()
        return (self.last_metric_read_ts < now - Agent.METRIC_READ_INTERVAL_SEC)
    
    def set_metric_read(self, now=None):
        self.last_metric_read_ts = now if now else time.time()

    def check_metric_upload_interval(self, now=None):
        if now is None:
            now = time.time()
        return (self.last_metric_upload_ts < now - Agent.METRIC_UPLOAD_INTERVAL_SEC)

    def set_metric_upload(self, now=None):
        self.last_metric_upload_ts = now if now else time.time()

    def tick(self, block=False, now=None):
        if now is None:
            now = time.time()
        if self.check_metric_upload_interval(now):
            if self._metric_store.has_unexported():
                metrics = self._metric_store.export()
                for metric in metrics:
                    self.uploader().upload_metric(metric)
                self.set_metric_upload(now)

        self.upload(block=False)


def _sha1(text, size=-1):
    sha1_hash = hashlib.sha1()
    sha1_hash.update(text.encode('utf-8'))
    return sha1_hash.hexdigest()[0:size]


def _uuid_sha1(size=-1):
    return _sha1(str(uuid.uuid4()), size)



import unittest
import logging
import sys
import os
from unittest.mock import patch, Mock

import graphsignal
from graphsignal.traces import Trace

logger = logging.getLogger('graphsignal')


class GraphsignalTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_check_and_set_arg_test(self):
        arg1 = graphsignal._check_and_set_arg('arg1', 'val1', is_str=True, required=True)
        self.assertEqual(arg1, 'val1')

        arg2 = graphsignal._check_and_set_arg('arg2', 1, is_int=True, required=True)
        self.assertEqual(arg2, 1)

        arg3 = graphsignal._check_and_set_arg('arg3', None, is_int=True, required=False)
        self.assertEqual(arg3, None)

        os.environ['GRAPHSIGNAL_ARG4'] = '2'
        arg4 = graphsignal._check_and_set_arg('arg4', None, is_int=True, required=False)
        self.assertEqual(arg4, 2)

        with self.assertRaises(ValueError):
            arg5 = graphsignal._check_and_set_arg('arg5', None, is_str=True, required=True)

        os.environ['GRAPHSIGNAL_ARG6'] = '10'
        arg6 = graphsignal._check_and_set_arg('arg6', None, is_int=True, required=True)
        self.assertEqual(arg6, 10)

        os.environ['GRAPHSIGNAL_ARG7'] = 'str'
        with self.assertRaises(ValueError):
            arg7 = graphsignal._check_and_set_arg('arg7', None, is_int=True, required=True)

        os.environ['GRAPHSIGNAL_ARG8'] = 'a=1, b = c '
        arg8 = graphsignal._check_and_set_arg('arg8', None, is_kv=True, required=True)
        self.assertEqual(arg8, {'a': '1', 'b': 'c'})

        os.environ['GRAPHSIGNAL_ARG9'] = 'a'
        with self.assertRaises(ValueError):
            arg9 = graphsignal._check_and_set_arg('arg9', None, is_kv=True, required=False)

    def test_configure(self):
        self.assertEqual(graphsignal._agent.api_key, 'k1')
        self.assertEqual(graphsignal._agent.debug_mode, True)

    @patch.object(Trace, '_stop', return_value=None)
    @patch.object(Trace, '_start', return_value=None)
    def test_trace_function(self, mocked_start, mocked_stop):
        @graphsignal.trace_function
        def test_func(p):
            return 1 + p

        ret = test_func(12)
        self.assertEqual(ret, 13)

        mocked_start.assert_called_once()
        mocked_stop.assert_called_once()

    @patch.object(Trace, '_stop', return_value=None)
    @patch.object(Trace, '_start', return_value=None)
    def test_trace_function_with_args(self, mocked_start, mocked_stop):
        @graphsignal.trace_function(endpoint='ep1', tags=dict(t1='v1'))
        def test_func(p):
            return 1 + p

        ret = test_func(12)
        self.assertEqual(ret, 13)

        mocked_start.assert_called_once()
        mocked_stop.assert_called_once()



import unittest
import logging
import sys
import os
import json
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint
import banana_dev as banana

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader
from graphsignal.traces import DEFAULT_OPTIONS
from graphsignal.recorders.banana_recorder import BananaRecorder

logger = logging.getLogger('graphsignal')


class BananaRecorderTest(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    async def asyncTearDown(self):
        graphsignal.shutdown()

    async def test_record(self):
        recorder = BananaRecorder()
        recorder.setup()
        proto = signals_pb2.Trace()
        context = {}
        recorder.on_trace_start(proto, context, DEFAULT_OPTIONS)
        recorder.on_trace_stop(proto, context, DEFAULT_OPTIONS)
        recorder.on_trace_read(proto, context, DEFAULT_OPTIONS)

        self.assertEqual(proto.frameworks[0].name, 'Banana Python SDK')

    @patch.object(Uploader, 'upload_trace')
    @patch.object(banana, 'run')
    async def test_trace_run(self, mocked_run, mocked_upload_trace):
        # mocking overrides autoinstrumentation, reinstrument
        recorder = BananaRecorder()
        recorder.setup()

        mocked_run.return_value = {
                "id": "12345678-1234-1234-1234-123456789012", 
                "message": "success", 
                "created": 1649712752, 
                "apiVersion": "26 Nov 2021", 
                "modelOutputs": [
                    {
                        "caption": "a baseball player throwing a ball"
                    },
                    {
                        "caption": "a baseball player throwing a bat"
                    }                    
                ]
            }

        response = banana.run(
            "api-key-1", 
            "model-key-1",
            ["test input 1", "test input 2"])

        recorder.shutdown()

        proto = mocked_upload_trace.call_args[0][0]

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(proto))

        self.assertEqual(proto.frameworks[0].name, 'Banana Python SDK')

        self.assertEqual(find_param(proto, 'model_key'), 'model-key-1')

        self.assertEqual(find_data_count(proto, 'model_inputs', 'byte_count'), 24.0)
        self.assertEqual(find_data_count(proto, 'model_inputs', 'element_count'), 2.0)
        self.assertEqual(find_data_count(proto, 'model_outputs', 'byte_count'), 65.0)
        self.assertEqual(find_data_count(proto, 'model_outputs', 'element_count'), 2.0)


def find_param(proto, name):
    for param in proto.params:
        if param.name == name:
            return param.value
    return None


def find_data_count(proto, data_name, count_name):
    for data_stats in proto.data_profile:
        if data_stats.data_name == data_name:
            for data_count in data_stats.counts:
                if data_count.name == count_name:
                    return data_count.count
    return None


from abc import ABC, abstractmethod

class BaseRecorder(ABC):
    def __init__(self):
        pass

    def setup(self):
        pass

    def shutdown(self):
        pass

    @abstractmethod
    def on_trace_start(self, proto, context, options):
        pass

    @abstractmethod
    def on_trace_stop(self, proto, context, options):
        pass

    @abstractmethod
    def on_trace_read(self, proto, context, options):
        pass

    @abstractmethod
    def on_metric_update(self):
        pass



import logging
import importlib
import functools

import graphsignal
from graphsignal.data.builtin_types import BuiltInTypesProfiler
from graphsignal.data.numpy_ndarray import NumpyNDArrayProfiler
from graphsignal.data.tf_tensor import TFTensorProfiler
from graphsignal.data.torch_tensor import TorchTensorProfiler
from graphsignal.data.base_data_profiler import DataStats

logger = logging.getLogger('graphsignal')


data_profilers = [
    BuiltInTypesProfiler(),
    NumpyNDArrayProfiler(),
    TorchTensorProfiler(),
    TFTensorProfiler()
]


def compute_data_stats(data):
    for dp in data_profilers:
        if dp.is_instance(data):
            return dp.compute_stats(data)

    return DataStats(type_name=type(data).__name__)


def encode_data_sample(data):
    for dp in data_profilers:
        if dp.is_instance(data):
            return dp.encode_sample(data)

    return None



import logging
import json

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.data.base_data_profiler import BaseDataProfiler, DataStats, DataSample

logger = logging.getLogger('graphsignal')


class NumpyNDArrayProfiler(BaseDataProfiler):
    def __init__(self):
        super().__init__()

    def is_instance(self, data):
        np = self.check_module('numpy')        
        return np is not None and isinstance(data, np.ndarray)

    def compute_stats(self, data):
        np = self.check_module('numpy')
        counts = {}
        counts['element_count'] = data.size
        counts['byte_count'] = data.nbytes
        counts['null_count'] = np.count_nonzero(data == None)
        if np.issubdtype(data.dtype, np.number):
            counts['nan_count'] = np.count_nonzero(np.isnan(data))
            counts['inf_count'] = np.count_nonzero(np.isinf(data))
        counts['zero_count'] = np.count_nonzero(data == 0)
        counts['negative_count'] = np.count_nonzero(data < 0)
        counts['positive_count'] = np.count_nonzero(data > 0)
        return DataStats(type_name='numpy.ndarray', shape=list(data.shape), counts=counts)

    def encode_sample(self, data):
        return DataSample(content_type='application/json', content_bytes=json.dumps(data.tolist()).encode('utf-8'))



import unittest
import logging
import sys
import os
import json
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint
import types
import openai

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader
from graphsignal.traces import DEFAULT_OPTIONS
from graphsignal.recorders.openai_recorder import OpenAIRecorder

logger = logging.getLogger('graphsignal')


class OpenAIRecorderTest(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    async def asyncTearDown(self):
        graphsignal.shutdown()

    async def test_record(self):
        recorder = OpenAIRecorder()
        recorder.setup()
        proto = signals_pb2.Trace()
        context = {}
        recorder.on_trace_start(proto, context, DEFAULT_OPTIONS)
        recorder.on_trace_stop(proto, context, DEFAULT_OPTIONS)
        recorder.on_trace_read(proto, context, DEFAULT_OPTIONS)
        self.assertEqual(proto.frameworks[0].name, 'OpenAI Python Library')

    @patch.object(Uploader, 'upload_trace')
    @patch.object(openai.Completion, 'create')
    async def test_completion_create(self, mocked_create, mocked_upload_trace):
        # mocking overrides autoinstrumentation, reinstrument
        recorder = OpenAIRecorder()
        recorder.setup()

        mocked_create.return_value = {
            "choices": [
                {
                    "finish_reason": "stop",
                    "index": 0,
                    "logprobs": None,
                    "text": "\n\ntest completion 1"
                },
                {
                    "finish_reason": "length",
                    "index": 1,
                    "logprobs": None,
                    "text": "\n\ntest completion 222"
                }
            ],
            "created": 1675070122,
            "id": "cmpl-id",
            "model": "text-davinci-003",
            "object": "text_completion",
            "usage": {
                "completion_tokens": 96,
                "prompt_tokens": 6,
                "total_tokens": 102
            }
        }

        response = openai.Completion.create(
            model="text-davinci-003", 
            prompt=['test prompt 1', 'test prompt 2'],
            temperature=0.1,
            top_p=1,
            max_tokens=1024,
            frequency_penalty=0,
            presence_penalty=0)

        recorder.shutdown()

        proto = mocked_upload_trace.call_args[0][0]

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(proto))

        self.assertEqual(find_param(proto, 'model'), 'text-davinci-003')
        self.assertEqual(find_param(proto, 'max_tokens'), '1024')
        self.assertEqual(find_param(proto, 'temperature'), '0.1')
        self.assertEqual(find_param(proto, 'top_p'), '1')
        self.assertEqual(find_param(proto, 'presence_penalty'), '0')
        self.assertEqual(find_param(proto, 'frequency_penalty'), '0')

        self.assertEqual(find_data_count(proto, 'prompt', 'byte_count'), 26.0)
        self.assertEqual(find_data_count(proto, 'prompt', 'element_count'), 2.0)
        self.assertEqual(find_data_count(proto, 'prompt', 'token_count'), 6.0)
        self.assertEqual(find_data_count(proto, 'completion', 'byte_count'), 40.0)
        self.assertEqual(find_data_count(proto, 'completion', 'element_count'), 2.0)
        self.assertEqual(find_data_count(proto, 'completion', 'token_count'), 96.0)
        self.assertEqual(find_data_count(proto, 'completion', 'finish_reason_stop'), 1.0)
        self.assertEqual(find_data_count(proto, 'completion', 'finish_reason_length'), 1.0)

    @patch.object(Uploader, 'upload_trace')
    @patch.object(openai.Completion, 'create')
    async def test_completion_create_stream(self, mocked_create, mocked_upload_trace):
        # mocking overrides autoinstrumentation, reinstrument
        recorder = OpenAIRecorder()
        recorder.setup()

        test_ret = [
            {
                "choices": [
                    {
                    "finish_reason": None,
                    "index": 1,
                    "logprobs": None,
                    "text": "abc"
                    }
                ],
                "created": 1676896808,
                "id": "cmpl-6lzkernKqOvF4ewZGhHlZ63HZJcbc",
                "model": "text-davinci-003",
                "object": "text_completion"
            },
            {
                "choices": [
                    {
                    "finish_reason": "stop",
                    "index": 1,
                    "logprobs": None,
                    "text": "\n"
                    }
                ],
                "created": 1676896808,
                "id": "cmpl-6lzkernKqOvF4ewZGhHlZ63HZJcbc",
                "model": "text-davinci-003",
                "object": "text_completion"
            }
        ]
        def test_ret_gen():
            for item in test_ret:
                yield item
        mocked_create.return_value = test_ret_gen()

        response = openai.Completion.create(
            model="text-davinci-003", 
            prompt=['count 1 to 3', 'generate 2 random letters'],
            temperature=0.1,
            top_p=1,
            max_tokens=1024,
            frequency_penalty=0,
            presence_penalty=0,
            stream=True)

        for r in response:
            pass

        recorder.shutdown()

        proto = mocked_upload_trace.call_args[0][0]

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(proto))
        self.assertEqual(find_data_count(proto, 'prompt', 'byte_count'), 37.0)
        self.assertEqual(find_data_count(proto, 'prompt', 'element_count'), 2.0)
        self.assertEqual(find_data_count(proto, 'completion', 'byte_count'), 4.0)
        self.assertEqual(find_data_count(proto, 'completion', 'element_count'), 2.0)
        self.assertEqual(find_data_count(proto, 'completion', 'finish_reason_stop'), 1.0)

    @patch.object(Uploader, 'upload_trace')
    @patch.object(openai.Completion, 'acreate')
    async def test_completion_acreate(self, mocked_create, mocked_upload_trace):
        # mocking overrides autoinstrumentation, reinstrument
        recorder = OpenAIRecorder()
        recorder.setup()

        mocked_create.return_value = {
            "choices": [
                {
                    "finish_reason": "stop",
                    "index": 0,
                    "logprobs": None,
                    "text": "\n\ntest completion 1"
                },
                {
                    "finish_reason": "stop",
                    "index": 1,
                    "logprobs": None,
                    "text": "\n\ntest completion 222"
                }
            ],
            "created": 1675070122,
            "id": "cmpl-id",
            "model": "text-davinci-003",
            "object": "text_completion",
            "usage": {
                "completion_tokens": 96,
                "prompt_tokens": 6,
                "total_tokens": 102
            }
        }

        response = await openai.Completion.acreate(
            model="text-davinci-003", 
            prompt=['test prompt 1', 'test prompt 2'],
            temperature=0.1,
            top_p=1,
            max_tokens=1024,
            frequency_penalty=0,
            presence_penalty=0)

        recorder.shutdown()

        proto = mocked_upload_trace.call_args[0][0]

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(proto))

        self.assertEqual(find_param(proto, 'model'), 'text-davinci-003')
        self.assertEqual(find_param(proto, 'max_tokens'), '1024')
        self.assertEqual(find_param(proto, 'temperature'), '0.1')
        self.assertEqual(find_param(proto, 'top_p'), '1')
        self.assertEqual(find_param(proto, 'presence_penalty'), '0')
        self.assertEqual(find_param(proto, 'frequency_penalty'), '0')

        self.assertEqual(find_data_count(proto, 'prompt', 'byte_count'), 26.0)
        self.assertEqual(find_data_count(proto, 'prompt', 'element_count'), 2.0)
        self.assertEqual(find_data_count(proto, 'prompt', 'token_count'), 6.0)
        self.assertEqual(find_data_count(proto, 'completion', 'byte_count'), 40.0)
        self.assertEqual(find_data_count(proto, 'completion', 'element_count'), 2.0)
        self.assertEqual(find_data_count(proto, 'completion', 'token_count'), 96.0)

    @patch.object(Uploader, 'upload_trace')
    @patch.object(openai.ChatCompletion, 'create')
    async def test_chat_completion_create(self, mocked_create, mocked_upload_trace):
        # mocking overrides autoinstrumentation, reinstrument
        recorder = OpenAIRecorder()
        recorder.setup()

        mocked_create.return_value = {
            "choices": [
                {
                    "finish_reason": "stop",
                    "index": 0,
                    "logprobs": None,
                    "message": {
                        "content": "\n\ntest completion 1"
                    }
                },
                {
                    "finish_reason": "length",
                    "index": 1,
                    "logprobs": None,
                    "message": {
                        "content": "\n\ntest completion 222"
                    }
                }
            ],
            "created": 1675070122,
            "id": "cmpl-id",
            "model": "gpt-3.5-turbo",
            "object": "text_completion",
            "usage": {
                "completion_tokens": 96,
                "prompt_tokens": 6,
                "total_tokens": 102
            }
        }

        response = openai.ChatCompletion.create(
            model="gpt-3.5-turbo", 
            messages=[
                {'role': 'system', 'content': 'test prompt 1'}, 
                {'role': 'system', 'content': 'test prompt 2'}],
            temperature=0.1,
            top_p=1,
            max_tokens=1024,
            frequency_penalty=0,
            presence_penalty=0)

        recorder.shutdown()

        proto = mocked_upload_trace.call_args[0][0]

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(proto))

        self.assertEqual(find_param(proto, 'model'), 'gpt-3.5-turbo')
        self.assertEqual(find_param(proto, 'max_tokens'), '1024')
        self.assertEqual(find_param(proto, 'temperature'), '0.1')
        self.assertEqual(find_param(proto, 'top_p'), '1')
        self.assertEqual(find_param(proto, 'presence_penalty'), '0')
        self.assertEqual(find_param(proto, 'frequency_penalty'), '0')

        self.assertEqual(find_data_count(proto, 'messages', 'byte_count'), 38.0)
        self.assertEqual(find_data_count(proto, 'messages', 'element_count'), 4.0)
        self.assertEqual(find_data_count(proto, 'messages', 'token_count'), 6.0)
        self.assertEqual(find_data_count(proto, 'completion', 'byte_count'), 40.0)
        self.assertEqual(find_data_count(proto, 'completion', 'element_count'), 2.0)
        self.assertEqual(find_data_count(proto, 'completion', 'token_count'), 96.0)
        self.assertEqual(find_data_count(proto, 'completion', 'finish_reason_stop'), 1.0)
        self.assertEqual(find_data_count(proto, 'completion', 'finish_reason_length'), 1.0)

    @patch.object(Uploader, 'upload_trace')
    @patch.object(openai.ChatCompletion, 'create')
    async def test_chat_completion_create_stream(self, mocked_create, mocked_upload_trace):
        # mocking overrides autoinstrumentation, reinstrument
        recorder = OpenAIRecorder()
        recorder.setup()

        test_ret = [
            {
                "choices": [
                    {
                        "finish_reason": None,
                        "index": 1,
                        "logprobs": None,
                        "delta": {
                            "content": "abc"
                        }
                    }
                ],
                "created": 1676896808,
                "id": "cmpl-6lzkernKqOvF4ewZGhHlZ63HZJcbc",
                "model": "gpt-3.5-turbo",
                "object": "chat.completion.chunk"
            },
            {
                "choices": [
                    {
                        "finish_reason": "stop",
                        "index": 1,
                        "logprobs": None,
                        "delta": {
                            "content": "\n"
                        }
                    }
                ],
                "created": 1676896808,
                "id": "cmpl-6lzkernKqOvF4ewZGhHlZ63HZJcbc",
                "model": "gpt-3.5-turbo",
                "object": "chat.completion.chunk"
            }
        ]
        def test_ret_gen():
            for item in test_ret:
                yield item
        mocked_create.return_value = test_ret_gen()

        response = openai.ChatCompletion.create(
            model="gpt-3.5-turbo", 
            messages=[
                {'role': 'system', 'content': 'count 1 to 3'}, 
                {'role': 'system', 'content': 'generate 2 random letters'}],            
            temperature=0.1,
            top_p=1,
            max_tokens=1024,
            frequency_penalty=0,
            presence_penalty=0,
            stream=True)

        for r in response:
            pass

        recorder.shutdown()

        proto = mocked_upload_trace.call_args[0][0]

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(proto))
        self.assertEqual(find_data_count(proto, 'messages', 'byte_count'), 49.0)
        self.assertEqual(find_data_count(proto, 'messages', 'element_count'), 4.0)
        self.assertEqual(find_data_count(proto, 'completion', 'byte_count'), 4.0)
        self.assertEqual(find_data_count(proto, 'completion', 'element_count'), 2.0)
        self.assertEqual(find_data_count(proto, 'completion', 'finish_reason_stop'), 1.0)

    @patch.object(Uploader, 'upload_trace')
    @patch.object(openai.Edit, 'create')
    async def test_edits_create(self, mocked_create, mocked_upload_trace):
        # mocking overrides autoinstrumentation, reinstrument
        recorder = OpenAIRecorder()
        recorder.setup()

        mocked_create.return_value = {
            "choices": [
                {
                    "index": 0,
                    "text": "test input 1\ntest instruction 1\n"
                }
            ],
            "created": 1675077592,
            "object": "edit",
            "usage": {
                "completion_tokens": 13,
                "prompt_tokens": 18,
                "total_tokens": 32
            }
        }

        response = openai.Edit.create(
            model="text-davinci-edit-001", 
            input='test input 1',
            instruction='test instruction 1',
            temperature=0.1,
            top_p=1)

        recorder.shutdown()

        proto = mocked_upload_trace.call_args[0][0]

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(proto))

        self.assertEqual(find_param(proto, 'model'), 'text-davinci-edit-001')
        self.assertEqual(find_param(proto, 'temperature'), '0.1')
        self.assertEqual(find_param(proto, 'top_p'), '1')

        self.assertEqual(find_data_count(proto, 'input', 'byte_count'), 12.0)
        self.assertEqual(find_data_count(proto, 'input', 'element_count'), 1.0)
        self.assertEqual(find_data_count(proto, 'instruction', 'byte_count'), 18.0)
        self.assertEqual(find_data_count(proto, 'instruction', 'element_count'), 1.0)
        self.assertEqual(find_data_count(proto, 'instruction', 'token_count'), 18.0)
        self.assertEqual(find_data_count(proto, 'edits', 'byte_count'), 32.0)
        self.assertEqual(find_data_count(proto, 'edits', 'element_count'), 1.0)
        self.assertEqual(find_data_count(proto, 'edits', 'token_count'), 13.0)

    @patch.object(Uploader, 'upload_trace')
    @patch.object(openai.Embedding, 'create')
    async def test_embedding_create(self, mocked_create, mocked_upload_trace):
        # mocking overrides autoinstrumentation, reinstrument
        recorder = OpenAIRecorder()
        recorder.setup()

        mocked_create.return_value = {
            "data": [
                {
                    "embedding": [
                        0.0007817253354005516,
                        -0.004567846190184355,
                        -0.011692041531205177
                    ],
                    "index": 0,
                    "object": "embedding"
                    },
                    {
                    "embedding": [
                        -0.008782976306974888,
                        -0.0015969815431162715,
                        -0.004394866060465574
                    ],
                    "index": 1,
                    "object": "embedding"
                    }
                ],
                "model": "text-embedding-ada-002-v2",
                "object": "list",
                "usage": {
                    "prompt_tokens": 8,
                    "total_tokens": 8
                }
            }

        response = openai.Embedding.create(
            engine="text-embedding-ada-002", 
            input=['test text 1', 'test text 2'])

        recorder.shutdown()

        proto = mocked_upload_trace.call_args[0][0]

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(proto))

        self.assertEqual(find_param(proto, 'engine'), 'text-embedding-ada-002')

        self.assertEqual(find_data_count(proto, 'input', 'byte_count'), 22.0)
        self.assertEqual(find_data_count(proto, 'input', 'element_count'), 2.0)
        self.assertEqual(find_data_count(proto, 'input', 'token_count'), 8.0)
        self.assertEqual(find_data_count(proto, 'embedding', 'byte_count'), 144.0)
        self.assertEqual(find_data_count(proto, 'embedding', 'element_count'), 6.0)

    @patch.object(Uploader, 'upload_trace')
    @patch.object(openai.Image, 'create')
    async def test_image_create(self, mocked_create, mocked_upload_trace):
        # mocking overrides autoinstrumentation, reinstrument
        recorder = OpenAIRecorder()
        recorder.setup()

        mocked_create.return_value = {
            "created": 1675079434,
            "data": [
                {
                    "b64_json": "iVBORw0KGgoAAA"
                }
            ]
        }

        response = openai.Image.create(
            prompt='test image',
            n=1,
            size='256x256',
            response_format='b64_json')

        recorder.shutdown()

        proto = mocked_upload_trace.call_args[0][0]

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(proto))

        self.assertEqual(find_param(proto, 'n'), '1')
        self.assertEqual(find_param(proto, 'size'), '256x256')
        self.assertEqual(find_param(proto, 'response_format'), 'b64_json')

        self.assertEqual(find_data_count(proto, 'image', 'byte_count'), 14.0)
        self.assertEqual(find_data_count(proto, 'image', 'element_count'), 1.0)

    @patch.object(Uploader, 'upload_trace')
    @patch.object(openai.Audio, 'transcribe')
    async def test_audio_transcription(self, mocked_create, mocked_upload_trace):
        # mocking overrides autoinstrumentation, reinstrument
        recorder = OpenAIRecorder()
        recorder.setup()

        mocked_create.return_value = {
            "text": 'some text'
        }

        import tempfile
        with tempfile.NamedTemporaryFile() as tmp_file:
            tmp_file.write('test content'.encode('utf-8'))
            tmp_file.flush()

            response = openai.Audio.transcribe(
                file=tmp_file,
                model='whisper-1',
                response_format='json',
                prompt='test prompt',
                temperature=0.1,
                language='en')

        recorder.shutdown()

        proto = mocked_upload_trace.call_args[0][0]

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(proto))

        self.assertEqual(find_param(proto, 'model'), 'whisper-1')
        self.assertEqual(find_param(proto, 'response_format'), 'json')
        self.assertEqual(find_param(proto, 'temperature'), '0.1')
        self.assertEqual(find_param(proto, 'language'), 'en')

        self.assertEqual(find_data_count(proto, 'file', 'byte_count'), 12.0)
        self.assertEqual(find_data_count(proto, 'prompt', 'byte_count'), 11.0)
        self.assertEqual(find_data_count(proto, 'prompt', 'element_count'), 1.0)


    @patch.object(Uploader, 'upload_trace')
    @patch.object(openai.Audio, 'translate')
    async def test_audio_translate(self, mocked_create, mocked_upload_trace):
        # mocking overrides autoinstrumentation, reinstrument
        recorder = OpenAIRecorder()
        recorder.setup()

        mocked_create.return_value = {
            "text": 'some text'
        }

        import tempfile
        with tempfile.NamedTemporaryFile() as tmp_file:
            tmp_file.write('test content'.encode('utf-8'))
            tmp_file.flush()

            response = openai.Audio.translate(
                file=tmp_file,
                model='whisper-1',
                response_format='json',
                prompt='test prompt',
                temperature=0.1)

        recorder.shutdown()

        proto = mocked_upload_trace.call_args[0][0]

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(proto))

        self.assertEqual(find_param(proto, 'model'), 'whisper-1')
        self.assertEqual(find_param(proto, 'response_format'), 'json')
        self.assertEqual(find_param(proto, 'temperature'), '0.1')

        self.assertEqual(find_data_count(proto, 'file', 'byte_count'), 12.0)
        self.assertEqual(find_data_count(proto, 'prompt', 'byte_count'), 11.0)
        self.assertEqual(find_data_count(proto, 'prompt', 'element_count'), 1.0)

    @patch.object(Uploader, 'upload_trace')
    @patch.object(openai.Moderation, 'create')
    async def test_moderation_create(self, mocked_create, mocked_upload_trace):
        # mocking overrides autoinstrumentation, reinstrument
        recorder = OpenAIRecorder()
        recorder.setup()

        mocked_create.return_value = {
            "id": "modr-id",
            "model": "text-moderation-004",
            "results": [
                {
                    "categories": {
                        "hate": False,
                        "hate/threatening": False,
                        "self-harm": False,
                        "sexual": False,
                        "sexual/minors": False,
                        "violence": False,
                        "violence/graphic": False
                    },
                    "category_scores": {
                        "hate": 1.5125541722227354e-05,
                        "hate/threatening": 4.6567969036459544e-08,
                        "self-harm": 4.265221065224978e-09,
                        "sexual": 0.00031956686871126294,
                        "sexual/minors": 4.4396051634976175e-06,
                        "violence": 6.315908080978261e-07,
                        "violence/graphic": 6.493833097920287e-07
                    },
                    "flagged": False
                }
            ]
        }

        response = openai.Moderation.create(
            model='text-moderation-latest',
            input='test text')

        recorder.shutdown()

        proto = mocked_upload_trace.call_args[0][0]

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(proto))

        self.assertEqual(find_param(proto, 'model'), 'text-moderation-latest')

        self.assertEqual(find_data_count(proto, 'input', 'byte_count'), 9.0)
        self.assertEqual(find_data_count(proto, 'input', 'element_count'), 1.0)


def find_param(proto, name):
    for param in proto.params:
        if param.name == name:
            return param.value
    return None


def find_data_count(proto, data_name, count_name):
    for data_stats in proto.data_profile:
        if data_stats.data_name == data_name:
            for data_count in data_stats.counts:
                if data_count.name == count_name:
                    return data_count.count
    return None


import logging
import sys
import time
import contextvars
import langchain
from typing import Any, Dict, List, Optional, Union
from langchain.callbacks.base import BaseCallbackHandler
from langchain.schema import AgentAction, AgentFinish, LLMResult

import graphsignal
from graphsignal.traces import TraceOptions
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.recorders.instrumentation import patch_method, unpatch_method
from graphsignal.proto_utils import parse_semver, compare_semver
from graphsignal.proto import signals_pb2
from graphsignal.proto_utils import add_framework_param, add_driver

logger = logging.getLogger('graphsignal')

trace_stack_var = contextvars.ContextVar('langchain_trace_stack', default=[])


def push_current_trace(trace):
    trace_stack_var.set(trace_stack_var.get() + [trace])


def pop_current_trace():
    trace_stack = trace_stack_var.get()
    if len(trace_stack) > 0:
        trace_stack_var.set(trace_stack[:-1])
        return trace_stack[-1]
    return None


def get_current_trace():
    trace_stack = trace_stack_var.get()
    if len(trace_stack) > 0:
        return trace_stack[-1]
    return None


class GraphsignalHandler(BaseCallbackHandler):
    def __init__(self) -> None:
        self._current_trace = None        

    @property
    def always_verbose(self) -> bool:
        return True

    def on_llm_start(
            self, serialized: Dict[str, Any], prompts: List[str], **kwargs: Any) -> None:
        name = 'langchain.llms.' + serialized.get('name', 'LLM')
        trace = graphsignal.start_trace(endpoint=name)
        push_current_trace(trace)
        if prompts:
            trace.set_data('prompts', prompts)

    def on_llm_new_token(self, token: str, **kwargs: Any) -> Any:
        pass

    def on_llm_end(self, response: LLMResult, **kwargs: Any) -> None:
        trace = pop_current_trace()
        if not trace:
            return
        trace.stop()

    def on_llm_error(
            self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any) -> None:
        trace = pop_current_trace()
        if not trace:
            return
        if isinstance(error, Exception):
            trace.set_exception(error)
        trace.stop()

    def on_chain_start(
            self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any) -> None:
        name = 'langchain.chains.' + serialized.get('name', 'Chain')
        trace = graphsignal.start_trace(endpoint=name)
        push_current_trace(trace)
        if inputs:
            trace.set_data('inputs', inputs)

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> None:
        trace = pop_current_trace()
        if not trace:
            return
        if outputs:
            trace.set_data('outputs', outputs)
        trace.stop()

    def on_chain_error(
            self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any) -> None:
        trace = pop_current_trace()
        if not trace:
            return
        if isinstance(error, Exception):
            trace.set_exception(error)
        trace.stop()

    def on_tool_start(
            self, serialized: Dict[str, Any], input_str: str, **kwargs: Any) -> None:
        name = 'langchain.agents.tools.' + serialized.get('name', 'Tool')
        trace = graphsignal.start_trace(endpoint=name)
        push_current_trace(trace)

    def on_tool_end(self, output: str, **kwargs: Any) -> None:
        trace = pop_current_trace()
        if not trace:
            return
        if output:
            trace.set_data('output', output)
        trace.stop()

    def on_tool_error(
            self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any) -> None:
        trace = pop_current_trace()
        if not trace:
            return
        if isinstance(error, Exception):
            trace.set_exception(error)
        trace.stop()

    def on_text(self, text: str, **kwargs: Optional[str]) -> None:
        pass

    def on_agent_action(self, action: AgentAction, **kwargs: Any) -> Any:
        pass

    def on_agent_finish(
            self, finish: AgentFinish, **kwargs: Any) -> None:
        pass


class LangChainRecorder(BaseRecorder):
    def __init__(self):
        self._framework = None
        self._handler = None

    def setup(self):
        if not graphsignal._agent.auto_instrument:
            return

        self._framework = signals_pb2.FrameworkInfo()
        self._framework.name = 'LangChain'

        if hasattr(langchain.callbacks, 'get_callback_manager'):
            self._handler = GraphsignalHandler()
            langchain.callbacks.get_callback_manager().add_handler(self._handler)

    def shutdown(self):
        if self._handler:
            langchain.callbacks.get_callback_manager().remove_handler(self._handler)
            self._handler = None

    def on_trace_start(self, proto, context, options):
        pass

    def on_trace_stop(self, proto, context, options):
        pass

    def on_trace_read(self, proto, context, options):
        if self._framework:
            proto.frameworks.append(self._framework)

    def on_metric_update(self):
        pass



import unittest
import logging
import sys
import os
import asyncio
from unittest.mock import patch, Mock

import graphsignal
from graphsignal.traces import Trace

logger = logging.getLogger('graphsignal')


class GraphsignalAsyncTest(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    async def asyncTearDown(self):
        graphsignal.shutdown()

    @patch.object(Trace, '_stop', return_value=None)
    @patch.object(Trace, '_start', return_value=None)
    async def test_start_trace(self, mocked_start, mocked_stop):
        async def test_func(p):
            return 1 + p

        with graphsignal.start_trace(endpoint='ep1'):
            ret = await test_func(12)
            self.assertEqual(ret, 13)

        mocked_start.assert_called_once()
        mocked_stop.assert_called_once()

    @patch.object(Trace, '_stop', return_value=None)
    @patch.object(Trace, '_start', return_value=None)
    async def test_trace_function(self, mocked_start, mocked_stop):
        @graphsignal.trace_function
        async def test_func(p):
            return 1 + p

        ret = await test_func(12)
        self.assertEqual(ret, 13)

        mocked_start.assert_called_once()
        mocked_stop.assert_called_once()

    @patch.object(Trace, '_stop', return_value=None)
    @patch.object(Trace, '_start', return_value=None)
    async def test_trace_function_with_args(self, mocked_start, mocked_stop):
        @graphsignal.trace_function(endpoint='ep1', tags=dict(t1='v1'))
        async def test_func(p):
            return 1 + p

        ret = await test_func(12)
        self.assertEqual(ret, 13)

        mocked_start.assert_called_once()
        mocked_stop.assert_called_once()



import logging
import contextvars
import time


logger = logging.getLogger('graphsignal')

span_stack_var = contextvars.ContextVar('graphsignal_span_stack', default=[])


def push_current_span(span):
    span_stack_var.set(span_stack_var.get() + [span])


def pop_current_span():
    span_stack = span_stack_var.get()
    if len(span_stack) > 0:
        span_stack_var.set(span_stack[:-1])
        return span_stack[-1]
    return None


def get_current_span():
    span_stack = span_stack_var.get()
    if len(span_stack) > 0:
        return span_stack[-1]
    return None


def is_current_span(span):
    return get_current_span() == span


class Span:
    __slots__ = [
        'endpoint',
        'trace_id',
        'parent_span',
        'root_span',
        'total_count',
        'in_context',
        'is_sampling'
    ]

    MAX_NESTED_SPANS = 250

    def __init__(self, endpoint):
        self.endpoint = endpoint
        self.trace_id = None
        self.root_span = None
        self.parent_span = None
        self.total_count = None
        self.in_context = False
        self.is_sampling = False

        parent_span = get_current_span()
        if not parent_span:
            self.init_as_root()
            push_current_span(self)
            self.in_context = True
        elif parent_span.can_add_child():
            self.init_as_child(parent_span)
            push_current_span(self)
            self.in_context = True

    def init_as_root(self):
        self.root_span = self
        self.total_count = 1
 
    def init_as_child(self, parent_span):
        self.parent_span = parent_span
        self.root_span = parent_span.root_span
        self.root_span.total_count += 1

    def set_trace_id(self, trace_id):
        self.trace_id = trace_id

    def set_sampling(self, is_sampling):
        self.is_sampling = is_sampling
    
    def is_root_sampling(self):
        return self.root_span and self.root_span.is_sampling

    def stop(self):
        if self.in_context:
            if is_current_span(self):
                pop_current_span()
            else:
                logger.error(f'Span.stop() called on a span that is not the current span {self.endpoint}')

    def can_add_child(self):
        return self.in_context and self.root_span.total_count < Span.MAX_NESTED_SPANS



from typing import Union, Any, Optional, Dict
import logging
import sys
import time
import traceback

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.data import compute_data_stats, encode_data_sample
from graphsignal.spans import get_current_span, Span

logger = logging.getLogger('graphsignal')


class TraceOptions:
    __slots__ = [
        'record_samples',
        'record_metrics',
        'enable_profiling'
    ]

    def __init__(self, 
            record_samples: bool = True,
            record_metrics: bool = True,
            enable_profiling: bool = False):
        self.record_samples = record_samples
        self.record_metrics = record_metrics
        self.enable_profiling = enable_profiling

DEFAULT_OPTIONS = TraceOptions()


class DataObject:
    __slots__ = [
        'name',
        'obj',
        'counts',
        'check_missing_values'
    ]

    def __init__(self, name, obj, counts=None, check_missing_values=False):
        self.name = name
        self.obj = obj
        self.counts = counts
        self.check_missing_values = check_missing_values


class Trace:
    MAX_RUN_TAGS = 10
    MAX_TRACE_TAGS = 10
    MAX_RUN_PARAMS = 10
    MAX_TRACE_PARAMS = 10
    MAX_DATA_OBJECTS = 10
    MAX_SAMPLE_BYTES = 32 * 1024
    MAX_NESTED_TRACES = 250

    __slots__ = [
        '_trace_sampler',
        '_lo_detector',
        '_mv_detector',
        '_agent',
        '_endpoint',
        '_tags',
        '_params',
        '_options',
        '_span',
        '_is_sampling',
        '_is_root',
        '_latency_ns',
        '_context',
        '_proto',
        '_is_started',
        '_is_stopped',
        '_start_counter',
        '_stop_counter',
        '_exc_info',
        '_data_objects',
        '_is_latency_outlier',
        '_has_missing_values'
    ]

    def __init__(self, endpoint, tags=None, options=None):
        self._is_started = False

        if not endpoint:
            logger.error('endpoint is required')
            return
        if tags is not None:
            if not isinstance(tags, dict):
                logger.error('tags must be dict')
                return
            if len(tags) > Trace.MAX_TRACE_TAGS:
                logger.error('too many tags (>{0})'.format(Trace.MAX_TRACE_TAGS))
                return

        self._endpoint = _sanitize_str(endpoint)
        if tags is not None:
            self._tags = dict(tags)
        else:
            self._tags = None
        self._params = None
        if options is None:
            self._options = DEFAULT_OPTIONS
        else:
            self._options = options
        self._agent = graphsignal._agent
        self._trace_sampler = None
        self._lo_detector = None
        self._mv_detector = None
        self._is_stopped = False
        self._span = None
        self._is_sampling = False
        self._is_root = False
        self._latency_ns = None
        self._start_counter = None
        self._stop_counter = None
        self._context = False
        self._proto = None
        self._exc_info = None
        self._data_objects = None
        self._is_latency_outlier = False
        self._has_missing_values = False

        try:
            self._start()
        except Exception as ex:
            self._is_stopped = True
            if self._is_sampling and self._is_root:
                self._trace_sampler.unlock()
            raise ex

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        if exc_info and exc_info[1] and isinstance(exc_info[1], Exception):
            self._exc_info = exc_info
        self.stop()
        return False

    def _init_sampling(self, sampling_type):
        self._is_sampling = True
        self._proto = self._agent.create_trace_proto()
        self._proto.sampling_type = sampling_type
        self._context = {}
        self._span.set_trace_id(self._proto.trace_id)
        self._span.set_sampling(True)

    def _start(self):
        if self._is_started:
            return
        if self._is_stopped:
            return

        if self._agent.debug_mode:
            logger.debug(f'Starting trace {self._endpoint}')

        self._trace_sampler = self._agent.trace_sampler(self._endpoint)
        self._lo_detector = self._agent.lo_detector(self._endpoint)
        self._mv_detector = self._agent.mv_detector()

        parent_span = get_current_span()
        self._span = Span(self._endpoint)
        self._is_root = not parent_span

        if self._options.record_samples:
            # sample if parent trace is being sampled or this trace is root
            if self._is_root:
                if self._trace_sampler.lock('samples'):
                    self._init_sampling(sampling_type=signals_pb2.Trace.SamplingType.RANDOM_SAMPLING)
            else:
                if parent_span.is_root_sampling() and parent_span.can_add_child():
                    self._init_sampling(sampling_type=signals_pb2.Trace.SamplingType.PARENT_SAMPLING)

            # emit start event
            if self._is_sampling and self._is_root:
                try:
                    self._agent.emit_trace_start(self._proto, self._context, self._options)
                except Exception as exc:
                    logger.error('Error in trace start event handlers', exc_info=True)
                    self._add_tracer_exception(exc)

        self._start_counter = time.perf_counter_ns()

        self._is_started = True

    def _measure(self) -> None:
        self._stop_counter = time.perf_counter_ns()

    def _stop(self) -> None:
        if not self._is_started:
            return
        if self._is_stopped:
            return
        self._is_stopped = True

        if self._agent.debug_mode:
            logger.debug(f'Stopping trace {self._endpoint}')

        if self._stop_counter is None:
            self._measure()
        self._latency_ns = self._stop_counter - self._start_counter

        now = time.time()
        end_us = int(now * 1e6)
        start_us = int(end_us - self._latency_ns / 1e3)
        now = int(now)

        # stop current span
        self._span.stop()

        # emit stop event
        if self._is_sampling and self._is_root:
            try:
                self._agent.emit_trace_stop(self._proto, self._context, self._options)
            except Exception as exc:
                logger.error('Error in trace stop event handlers', exc_info=True)
                self._add_tracer_exception(exc)

        trace_tags = self._trace_tags()

        # if exception, but the trace is not being recorded, try to start tracing
        if self._options.record_samples:
            if not self._is_sampling and self._exc_info and self._exc_info[0]:
                if self._trace_sampler.lock('exceptions'):
                    self._init_sampling(sampling_type=signals_pb2.Trace.SamplingType.ERROR_SAMPLING)

        # check for outliers
        if self._options.record_samples:
            self._is_latency_outlier = self._lo_detector.detect(self._latency_ns / 1e9)
            if not self._is_sampling and self._is_latency_outlier:
                if self._trace_sampler.lock('latency-outliers'):
                    self._init_sampling(sampling_type=signals_pb2.Trace.SamplingType.ERROR_SAMPLING)
        self._lo_detector.update(self._latency_ns / 1e9)

        # update RED metrics
        if self._options.record_metrics:
            self._agent.metric_store().update_histogram(
                scope='performance', name='latency', tags=trace_tags, value=self._latency_ns, update_ts=now, is_time=True)
            self._agent.metric_store().inc_counter(
                scope='performance', name='call_count', tags=trace_tags, value=1, update_ts=now)
            if self._exc_info and self._exc_info[0]:
                self._agent.metric_store().inc_counter(
                    scope='performance', name='exception_count', tags=trace_tags, value=1, update_ts=now)
                self.set_tag('exception', self._exc_info[0].__name__)

        # compute data statistics
        data_stats = None
        if self._options.record_metrics and self._data_objects is not None:
            data_stats = {}
            for data_obj in self._data_objects.values():
                try:
                    stats = compute_data_stats(data_obj.obj)
                    if not stats:
                        continue
                    data_stats[data_obj.name] = stats

                    # add/update extra counts to computed counts
                    if data_obj.counts is not None:
                        stats.counts.update(data_obj.counts)

                    # check missing values
                    if data_obj.check_missing_values:
                        if self._mv_detector.detect(data_obj.name, stats.counts):
                            self._has_missing_values = True

                    # update data metrics
                    data_tags = trace_tags.copy()
                    data_tags['data'] = data_obj.name
                    for count_name, count in stats.counts.items():
                        self._agent.metric_store().inc_counter(
                            scope='data', name=count_name, tags=data_tags, value=count, update_ts=now)
                except Exception as exc:
                    logger.error('Error computing data stats', exc_info=True)
                    self._add_tracer_exception(exc)

        # if missing values detected, but the trace is not being recorded, try to start tracing
        if self._options.record_samples:
            if not self._is_sampling and self._has_missing_values:
                if self._trace_sampler.lock('missing-values'):
                    self._init_sampling(sampling_type=signals_pb2.Trace.SamplingType.ERROR_SAMPLING)

        # emit read event
        if self._is_sampling and self._is_root:
            try:
                self._agent.emit_trace_read(self._proto, self._context, self._options)
            except Exception as exc:
                logger.error('Error in trace read event handlers', exc_info=True)
                self._add_tracer_exception(exc)

        # update recorder metrics
        if self._options.record_metrics and self._agent.check_metric_read_interval(now):
            try:
                self._agent.emit_metric_update()
                self._agent.set_metric_read(now)
            except Exception as exc:
                logger.error('Error in trace read event handlers', exc_info=True)
                self._add_tracer_exception(exc)

        # fill and upload trace
        if self._is_sampling:
            # copy data to trace proto
            self._proto.start_us = start_us
            self._proto.end_us = end_us
            if self._is_root:
                self._proto.labels.insert(0, 'root')
            if self._exc_info and self._exc_info[0]:
                self._proto.labels.append('exception')
            if self._is_latency_outlier:
                self._proto.labels.append('latency-outlier')
            if self._has_missing_values:
                self._proto.labels.append('missing-values')
            self._proto.process_usage.start_ms = self._agent._process_start_ms

            # copy span
            self._proto.span.start_ns = self._start_counter
            self._proto.span.end_ns = self._stop_counter
            if self._span.parent_span and self._span.parent_span.trace_id:
                self._proto.span.parent_trace_id = self._span.parent_span.trace_id
            if self._span.root_span and self._span.root_span.trace_id:
                self._proto.span.root_trace_id = self._span.root_span.trace_id

            # copy tags
            for key, value in trace_tags.items():
                tag = self._proto.tags.add()
                tag.key = _sanitize_str(key)
                tag.value = _sanitize_str(value)

            # copy params
            if self._params is not None:
                for name, value in self._params.items():
                    param = self._proto.params.add()
                    param.name = _sanitize_str(name)
                    param.value = _sanitize_str(value)

            # copy exception
            if self._exc_info and self._exc_info[0]:
                exception_proto = self._proto.exceptions.add()
                if self._exc_info[0] and hasattr(self._exc_info[0], '__name__'):
                    exception_proto.exc_type = str(self._exc_info[0].__name__)
                if self._exc_info[1]:
                    exception_proto.message = str(self._exc_info[1])
                if self._exc_info[2]:
                    frames = traceback.format_tb(self._exc_info[2])
                    if len(frames) > 0:
                        exception_proto.stack_trace = ''.join(frames)

            # copy data stats
            if data_stats is not None:
                for data_name, stats in data_stats.items():
                    data_stats_proto = self._proto.data_profile.add()
                    data_stats_proto.data_name = data_name
                    if stats.type_name:
                        data_stats_proto.data_type = stats.type_name
                    if stats.shape:
                        data_stats_proto.shape[:] = stats.shape
                    for counter_name, count in stats.counts.items():
                        if count > 0:
                            dc = data_stats_proto.counts.add()
                            dc.name = counter_name
                            dc.count = count
                    if self._agent.record_data_samples:
                        sample = encode_data_sample(self._data_objects[data_name].obj)
                        if sample is not None and len(sample.content_bytes) <= Trace.MAX_SAMPLE_BYTES:
                            sample_proto = self._proto.data_samples.add()
                            sample_proto.data_name = data_name
                            sample_proto.content_type = sample.content_type
                            sample_proto.content_bytes = sample.content_bytes

            # queue trace proto for upload
            self._agent.uploader().upload_trace(self._proto)

        # trigger upload
        if self._is_root:
            self._agent.tick(now)

    def measure(self) -> None:
        if not self._is_stopped:
            self._measure()

    def stop(self) -> None:
        try:
            self._stop()
        finally:
            self._is_stopped = True
            if self._is_sampling and self._is_root:
                self._trace_sampler.unlock()

    def is_sampling(self):
        return self._is_sampling

    def set_tag(self, key: str, value: str) -> None:
        if not key:
            logger.error('set_tag: key must be provided')
            return
        if value is None:
            logger.error('set_tag: value must be provided')
            return

        if self._tags is None:
            self._tags = {}

        if len(self._tags) > Trace.MAX_TRACE_TAGS:
            logger.error('set_tag: too many tags (>{0})'.format(Trace.MAX_TRACE_TAGS))
            return

        self._tags[key] = value

    def set_param(self, name: str, value: str) -> None:
        if not name:
            logger.error('set_param: name must be provided')
            return

        if self._params is None:
            self._params = {}

        if len(self._params) > Trace.MAX_TRACE_PARAMS:
            logger.error('set_param: too many params (>{0})'.format(Trace.MAX_TRACE_PARAMS))
            return

        self._params[name] = value

    def set_exception(self, exc: Optional[Exception] = None, exc_info: Optional[bool] = None) -> None:
        if exc is not None and not isinstance(exc, Exception):
            logger.error('set_exception: exc must be instance of Exception')
            return

        if exc_info is not None and not isinstance(exc_info, bool):
            logger.error('set_exception: exc_info must be bool')
            return

        if exc:
            self._exc_info = (exc.__class__, str(exc), exc.__traceback__)
        elif exc_info == True:
            self._exc_info = sys.exc_info()

    def set_data(self, 
            name: str, 
            obj: Any, 
            counts: Optional[Dict[str, int]] = None, 
            check_missing_values: Optional[bool] = False) -> None:
        if self._data_objects is None:
            self._data_objects = {}

        if name and not isinstance(name, str):
            logger.error('set_data: name must be string')
            return

        if counts and not isinstance(counts, dict):
            logger.error('append_data: name must be dict')
            return

        if len(self._data_objects) > Trace.MAX_DATA_OBJECTS:
            logger.error('set_data: too many data objects (>{0})'.format(Trace.MAX_DATA_OBJECTS))
            return

        self._data_objects[name] = DataObject(
            name=name, obj=obj, counts=counts, check_missing_values=check_missing_values)

    def append_data(self, 
            name: str, 
            obj: Any, 
            counts: Optional[Dict[str, int]] = None, 
            check_missing_values: Optional[bool] = False) -> None:
        if self._data_objects is None:
            self._data_objects = {}

        if name and not isinstance(name, str):
            logger.error('append_data: name must be string')
            return

        if counts and not isinstance(counts, dict):
            logger.error('append_data: name must be dict')
            return

        if len(self._data_objects) > Trace.MAX_DATA_OBJECTS:
            logger.error('append_data: too many data objects (>{0})'.format(Trace.MAX_DATA_OBJECTS))
            return

        if name in self._data_objects:
            data_obj = self._data_objects[name]
            data_obj.obj += obj
            if counts:
                if data_obj.counts is None:
                    data_obj.counts = {}
                for count_name, value in counts.items():
                    if count_name in data_obj.counts:
                        data_obj.counts[count_name] += value
                    else:
                        data_obj.counts[count_name] = value
        else:
            self._data_objects[name] = DataObject(
                name=name, obj=obj, counts=counts, check_missing_values=check_missing_values)

    def _add_tracer_exception(self, exc):
        if not self._is_sampling:
            return

        tracer_error = self._proto.tracer_errors.add()
        tracer_error.message = str(exc)
        if exc.__traceback__:
            frames = traceback.format_tb(exc.__traceback__)
            if len(frames) > 0:
                tracer_error.stack_trace = ''.join(frames)

    def _trace_tags(self, extra_tags=None):
        trace_tags = {
            'deployment': self._agent.deployment, 
            'endpoint': self._endpoint}
        if self._agent.hostname:
            trace_tags['hostname'] = self._agent.hostname
        if self._agent.tags is not None:
            trace_tags.update(self._agent.tags)
        context_tags = self._agent.context_tags.get()
        if len(context_tags) > 0:
            trace_tags.update(context_tags)
        if self._tags is not None:
            trace_tags.update(self._tags)
        if extra_tags is not None:
            trace_tags.update(extra_tags)
        return trace_tags

    def repr(self):
        return 'Trace({0})'.format(self._endpoint)


def _sanitize_str(val, max_len=50):
    return str(val)[:max_len]


import unittest
import logging
import sys
import time
from unittest.mock import patch, Mock
import random

import graphsignal
from graphsignal.detectors.latency_outlier_detector import LatencyOutlierDetector

logger = logging.getLogger('graphsignal')

class LatencyOutlierDetectorTest(unittest.TestCase):
    def setUp(self):
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_detect(self):
        detector = LatencyOutlierDetector()

        for i in range(500):
            val = random.randint(100, 200)
            self.assertFalse(detector.detect(val))
            detector.update(val)
        self.assertTrue(detector.detect(1000))


import unittest
import logging
import sys
import os
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader

logger = logging.getLogger('graphsignal')

class AgentTest(unittest.TestCase):
    def setUp(self):
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_create_trace_proto(self):
        proto = graphsignal._agent.create_trace_proto()
        self.assertTrue(proto.tracer_info.version.major > 0 or proto.tracer_info.version.minor > 0)

    @patch.object(Uploader, 'upload_metric')
    def test_shutdown_upload(self, mocked_upload_metric):
        graphsignal._agent.metric_store().set_gauge(scope='s1', name='n1', tags={}, value=1, update_ts=1)
        graphsignal.shutdown()

        proto = mocked_upload_metric.call_args[0][0]

        self.assertEqual(proto.scope, 's1')
        self.assertEqual(proto.name, 'n1')



import logging
import sys
import time
import torch
import deepspeed
import deepspeed.comm as dist

import graphsignal
from graphsignal.traces import TraceOptions
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.recorders.instrumentation import patch_method
from graphsignal.proto_utils import parse_semver, compare_semver
from graphsignal.proto import signals_pb2
from graphsignal.proto_utils import add_framework_param, add_driver

logger = logging.getLogger('graphsignal')

class DeepSpeedRecorder(BaseRecorder):
    COLLECTIVE_OPS = [
        'broadcast',
        'all_gather',
        'reduce_scatter_base',
        'all_gather_base',
        'all_to_all_single',
        'send',
        'recv',
        'isend',
        'irecv',
        'gather',
        'scatter',
        'barrier',
        'monitored_barrier',
        'reduce',
        'reduce_scatter',
        'all_reduce'
    ]

    def __init__(self):
        self._is_initialized = False
        self._framework = None
        self._rank = None
        self._local_rank = None
        self._world_size = None
        self._is_sampling = False
        self._op_profile = None

    def setup(self):
        self._framework = signals_pb2.FrameworkInfo()
        self._framework.name = 'DeepSpeed'
        parse_semver(self._framework.version, deepspeed.__version__)

        if dist.is_initialized():
            add_framework_param(self._framework, 'deepspeed.comm.get_world_size', dist.get_world_size())
            add_framework_param(self._framework, 'deepspeed.comm.get_rank', dist.get_rank())
            add_framework_param(self._framework, 'deepspeed.comm.get_local_rank', dist.get_local_rank())

            self._rank = dist.get_rank()
            self._local_rank = dist.get_local_rank()
            self._world_size = dist.get_world_size()

            if compare_semver(self._framework.version, (0, 7, 0)) < 1:
                logger.debug('DeepSpeed tracing is only supported for >= 0.7.0.')
                return

            # auto-instrument collective operations to be traced as standalone endpoints
            for op_name in DeepSpeedRecorder.COLLECTIVE_OPS:
                self._instrument_op(op_name)

            self._is_initialized = True

    def _instrument_op(self, op_name):
        def before_op(args, kwargs):
            if self._is_sampling:
                return time.perf_counter_ns()
            return None

        def after_op(args, kwargs, ret, exc, start_counter):
            if self._is_sampling and start_counter is not None:
                stop_counter = time.perf_counter_ns()
                latency_us = int((stop_counter - start_counter) / 1e3)
                data_size = self._get_data_size(op_name, args, kwargs)

                if not self._op_profile:
                    self._op_profile = {}
                if op_name not in self._op_profile:
                    self._op_profile[op_name] = {
                        'count': 1,
                        'total_time_us': latency_us,
                        'total_data_size': data_size
                    }
                else:
                    op_stats = self._op_profile[op_name]
                    op_stats['count'] += 1
                    op_stats['total_time_us'] += latency_us
                    op_stats['total_data_size'] += data_size

        if not patch_method(dist, op_name, before_func=before_op, after_func=after_op):
            logger.debug('Cannot instrument DeepSpeed communications logger.')

    def on_trace_start(self, proto, context, options):
        if not self._is_initialized:
            return

        self._is_sampling = True

    def on_trace_stop(self, proto, context, options):
        if not self._is_initialized:
            return

        self._is_sampling = False

    def on_trace_read(self, proto, context, options):
        if not self._is_initialized:
            return

        if self._framework:
            proto.frameworks.append(self._framework)
        if self._rank is not None:
            proto.process_usage.rank = self._rank
            proto.process_usage.has_rank = True            
        if self._local_rank is not None:
            proto.process_usage.local_rank = self._local_rank
            proto.process_usage.has_local_rank = True

        if self._op_profile:
            for name, stats in self._op_profile.items():
                op_stats = proto.op_profile.add()
                op_stats.op_type = signals_pb2.OpStats.OpType.COLLECTIVE_OP
                op_stats.op_name = name
                op_stats.count = stats['count']
                op_stats.host_time_ns = int(stats['total_time_us'] * 1e3)
                op_stats.data_size = stats['total_data_size']
                if op_stats.host_time_ns > 0:
                    op_stats.data_per_sec = op_stats.data_size / (op_stats.host_time_ns / 1e9)

            self._op_profile = None

    def on_metric_update(self):
        pass

    def _get_data_size(self, op_name, args, kwargs):
        tensor_arg = None

        if len(args) > 0:
            tensor_arg = args[0]
        elif len(kwargs) > 0:
            if 'tensor' in kwargs:
                tensor_arg = kwargs['tensor']
            elif 'output_tensor' in kwargs:
                tensor_arg = kwargs['output_tensor']
            elif 'output' in kwargs:
                tensor_arg = kwargs['output']
            elif 'tensor_list' in kwargs:
                tensor_arg = kwargs['tensor_list']

        msg_size = 0
        if isinstance(tensor_arg, torch.Tensor):
            msg_size = tensor_arg.element_size() * tensor_arg.nelement()
        elif isinstance(tensor_arg, list) and len(tensor_arg) > 0 and isinstance(tensor_arg[0], torch.Tensor):
            msg_size = sum(el.element_size() * el.nelement() for el in tensor_arg)
        else:
            return 0

        if op_name in ("all_gather", "all_gather_base", "reduce_scatter", "reduce_scatter_base"):
            return msg_size * self._world_size
        elif op_name == "all_reduce":
            return msg_size * 2
        else:
            return msg_size




import unittest
import logging
import sys
import os
import json
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint
import torch
import torch.multiprocessing as mp
import deepspeed

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader
from graphsignal.traces import DEFAULT_OPTIONS
from graphsignal.recorders.deepspeed_recorder import DeepSpeedRecorder

logger = logging.getLogger('graphsignal')


class DeepSpeedRecorderTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_record(self):
        import torch
        if not torch.cuda.is_available() or torch.cuda.device_count() < 2:
            return

        mp.set_start_method("spawn")
        p2 = mp.Process(target=init_process, args=(1, 2))
        p2.start()

        init_process(0, 2)

        recorder = DeepSpeedRecorder()
        recorder.setup()
        proto = signals_pb2.Trace()
        context = {}
        recorder.on_trace_start(proto, context, DEFAULT_OPTIONS)

        tensor = torch.zeros(1)
        tensor += 1
        deepspeed.comm.send(tensor=tensor.cuda(), dst=1)

        recorder.on_trace_stop(proto, context, DEFAULT_OPTIONS)
        recorder.on_trace_read(proto, context, DEFAULT_OPTIONS)

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(proto))

        self.assertEqual(proto.frameworks[0].name, 'DeepSpeed')

        self.assertEqual(proto.frameworks[0].params[0].name, 'deepspeed.comm.get_world_size')
        self.assertEqual(proto.frameworks[0].params[0].value, str(2))

        self.assertEqual(proto.op_profile[0].op_type, signals_pb2.OpStats.OpType.COLLECTIVE_OP)
        self.assertEqual(proto.op_profile[0].op_name, 'send')
        self.assertEqual(proto.op_profile[0].count, 1)
        self.assertTrue(proto.op_profile[0].host_time_ns > 0)
        self.assertTrue(proto.op_profile[0].data_size > 0)
        self.assertTrue(proto.op_profile[0].data_per_sec > 0)

        p2.terminate()


def init_process(rank, size):
    logger.info('Initializing process %d (%d)', rank, size)

    os.environ["RANK"] = str(rank)
    os.environ["LOCAL_RANK"] = str(rank)
    os.environ["WORLD_SIZE"] = str(size)
    os.environ["MASTER_ADDR"] = "127.0.0.1"
    os.environ["MASTER_PORT"] = "29501"

    deepspeed.init_distributed()

    torch.cuda.set_device(rank)

    if rank == 1:
        tensor = torch.zeros(1)
        deepspeed.comm.recv(tensor=tensor.cuda(), src=0)


def start_processes(size):
    processes = []
    mp.set_start_method("spawn")
    for rank in range(size):
        p = mp.Process(target=init_process, args=(rank, size))
        p.start()
        processes.append(p)

    for p in processes:
        p.join()


import logging
import contextvars
import time


logger = logging.getLogger('graphsignal')


class LatencyOutlierDetector:
    MIN_VALUES = 100
    MAX_VALUES = 1000

    def __init__(self):
        self.reset()

    def reset(self):
        self._sum = 0
        self._sum2 = 0
        self._count = 0

    def update(self, value):
        self._sum += value
        self._sum2 += value * value
        self._count += 1

    def detect(self, value):
        if self._count < self.MIN_VALUES:
            return False
        elif self._count > self.MAX_VALUES:
            self.reset()
            return False
        mean = self._sum / self._count
        std = (self._sum2 / self._count - mean * mean) ** 0.5
        return value > mean + 6 * std



import logging
import os
import sys
import time

import graphsignal
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.proto import signals_pb2
from graphsignal.vendor.pynvml.pynvml import *

logger = logging.getLogger('graphsignal')

class NVMLRecorder(BaseRecorder):
    MIN_SAMPLE_READ_INTERVAL_US = int(10 * 1e6)

    def __init__(self):
        self._is_initialized = False
        self._setup_us = None
        self._last_nvlink_throughput_data_tx = {}
        self._last_nvlink_throughput_data_rx = {}

    def setup(self):
        try:
            nvmlInit()
            self._is_initialized = True
            logger.debug('Initialized NVML')
        except BaseException:
            logger.debug('Error initializing NVML, skipping GPU usage')

        self._setup_us = int(time.time() * 1e6)

    def shutdown(self):
        if not self._is_initialized:
            return

        try:
            nvmlShutdown()
            self._is_initialized = False
        except BaseException:
            logger.error('Error shutting down NVML', exc_info=True)

    def on_trace_start(self, proto, context, options):
        pass

    def on_trace_stop(self, proto, context, options):
        pass

    def on_trace_read(self, proto, context, options):
        if not self._is_initialized:
            return

        now_us = int(time.time() * 1e6)

        device_count = nvmlDeviceGetCount()

        proto.node_usage.num_devices = device_count

        if device_count > 0:
            try:
                version = nvmlSystemGetCudaDriverVersion_v2()
                if version:
                    driver_info = proto.node_usage.drivers.add()
                    driver_info.name = 'CUDA'
                    driver_info.version = _format_version(version)
            except NVMLError as err:
                _log_nvml_error(err)

        for idx in range(0, device_count):
            try:
                handle = nvmlDeviceGetHandleByIndex(idx)
            except NVMLError as err:
                _log_nvml_error(err)
                continue

            device_usage = proto.device_usage.add()
            device_usage.device_type = signals_pb2.DeviceUsage.DeviceType.GPU_DEVICE
            device_usage.device_idx = idx

            try:
                pci_info = nvmlDeviceGetPciInfo(handle)
                device_usage.device_id = pci_info.busId
            except NVMLError as err:
                _log_nvml_error(err)

            try:
                device_usage.device_name = nvmlDeviceGetName(handle)
            except NVMLError as err:
                _log_nvml_error(err)

            try:
                arch = nvmlDeviceGetArchitecture(handle)
                if arch == NVML_DEVICE_ARCH_KEPLER:
                    device_usage.architecture = 'Kepler'
                elif arch == NVML_DEVICE_ARCH_MAXWELL:
                    device_usage.architecture = 'Maxwell'
                elif arch == NVML_DEVICE_ARCH_PASCAL:
                    device_usage.architecture = 'Pascal'
                elif arch == NVML_DEVICE_ARCH_VOLTA:
                    device_usage.architecture = 'Volta'
                elif arch == NVML_DEVICE_ARCH_TURING:
                    device_usage.architecture = 'Turing'
                elif arch == NVML_DEVICE_ARCH_AMPERE:
                    device_usage.architecture = 'Ampere'
                elif arch == NVML_DEVICE_ARCH_ADA:
                    device_usage.architecture = 'Ada'
                elif arch == NVML_DEVICE_ARCH_HOPPER:
                    device_usage.architecture = 'Hopper'
            except NVMLError as err:
                _log_nvml_error(err)

            try:
                cc_major, cc_minor = nvmlDeviceGetCudaComputeCapability(handle)
                device_usage.compute_capability.major = cc_major
                device_usage.compute_capability.minor = cc_minor
            except NVMLError as err:
                _log_nvml_error(err)

            try:
                mem_info = nvmlDeviceGetMemoryInfo_v2(handle)
                device_usage.mem_total = mem_info.total
                device_usage.mem_used = mem_info.used
                device_usage.mem_free = mem_info.total - mem_info.used
                device_usage.mem_reserved = mem_info.reserved
            except NVMLError as err:
                try:
                    mem_info = nvmlDeviceGetMemoryInfo(handle)
                    device_usage.mem_total = mem_info.total
                    device_usage.mem_used = mem_info.used
                    device_usage.mem_free = mem_info.total - mem_info.used
                except NVMLError as err:
                    _log_nvml_error(err)

            seen_pids = set()
            process_info_fns = [
                nvmlDeviceGetComputeRunningProcesses, 
                nvmlDeviceGetMPSComputeRunningProcesses, 
                nvmlDeviceGetGraphicsRunningProcesses]
            for process_info_fn in process_info_fns:
                try:
                    process_infos = process_info_fn(handle)
                    for process_info in process_infos:
                        if process_info.pid not in seen_pids:
                            seen_pids.add(process_info.pid)
                            device_process_usage = device_usage.processes.add()
                            device_process_usage.pid = process_info.pid
                            device_process_usage.compute_instance_id = process_info.computeInstanceId
                            device_process_usage.gpu_instance_id = process_info.gpuInstanceId
                            if process_info.usedGpuMemory:
                                device_process_usage.mem_used = process_info.usedGpuMemory
                except NVMLError as err:
                    _log_nvml_error(err)

            try:
                last_read_us = max(
                    int(self._setup_us),
                    now_us - NVMLRecorder.MIN_SAMPLE_READ_INTERVAL_US)

                sample_value_type, gpu_samples = nvmlDeviceGetSamples(handle, NVML_GPU_UTILIZATION_SAMPLES, last_read_us)
                device_usage.gpu_utilization_percent = _avg_sample_value(sample_value_type, gpu_samples)

                sample_value_type, mem_samples = nvmlDeviceGetSamples(handle, NVML_MEMORY_UTILIZATION_SAMPLES, last_read_us)
                device_usage.mem_access_percent = _avg_sample_value(sample_value_type, mem_samples)
            except NVMLError as err:
                _log_nvml_error(err)

            try:
                device_usage.pcie_throughput_tx = nvmlDeviceGetPcieThroughput(
                    handle, NVML_PCIE_UTIL_TX_BYTES)
                device_usage.pcie_throughput_rx = nvmlDeviceGetPcieThroughput(
                    handle, NVML_PCIE_UTIL_RX_BYTES)
            except NVMLError as err:
                _log_nvml_error(err)

            try:
                nvlink_throughput_data_tx = nvmlDeviceGetFieldValues(handle, [NVML_FI_DEV_NVLINK_THROUGHPUT_DATA_TX])[0]
                if nvlink_throughput_data_tx.nvmlReturn == NVML_SUCCESS:
                    value = _nvml_value(nvlink_throughput_data_tx.valueType, nvlink_throughput_data_tx.value)
                    if idx in self._last_nvlink_throughput_data_tx:
                        last_data = self._last_nvlink_throughput_data_tx[idx]
                        last_value = _nvml_value(last_data.valueType, last_data.value)
                        interval_us = nvlink_throughput_data_tx.timestamp - last_data.timestamp
                        if interval_us > 0:
                            device_usage.nvlink_throughput_data_tx_kibs = (value - last_value) / (interval_us * 1e6)
                    self._last_nvlink_throughput_data_tx[idx] = nvlink_throughput_data_tx

                nvlink_throughput_data_rx = nvmlDeviceGetFieldValues(handle, [NVML_FI_DEV_NVLINK_THROUGHPUT_DATA_RX])[0]
                if nvlink_throughput_data_rx.nvmlReturn == NVML_SUCCESS:
                    value = _nvml_value(nvlink_throughput_data_rx.valueType, nvlink_throughput_data_rx.value)
                    if idx in self._last_nvlink_throughput_data_rx:
                        last_data = self._last_nvlink_throughput_data_rx[idx]
                        last_value = _nvml_value(last_data.valueType, last_data.value)
                        interval_us = nvlink_throughput_data_rx.timestamp - last_data.timestamp
                        if interval_us > 0:
                            device_usage.nvlink_throughput_data_rx_kibs = (value - last_value) / (interval_us * 1e6)
                    self._last_nvlink_throughput_data_rx[idx] = nvlink_throughput_data_rx
            except NVMLError as err:
                log_nvml_error(err)

            try:
                device_usage.gpu_temp_c = nvmlDeviceGetTemperature(
                    handle, NVML_TEMPERATURE_GPU)
            except NVMLError as err:
                _log_nvml_error(err)

            try:
                device_usage.power_usage_w = nvmlDeviceGetPowerUsage(handle) / 1000.0
            except NVMLError as err:
                _log_nvml_error(err)

            try:
                device_usage.fan_speed_percent = nvmlDeviceGetFanSpeed(handle)
            except NVMLError as err:
                _log_nvml_error(err)

    def on_metric_update(self):
        now = int(time.time())
        proto = signals_pb2.Trace()
        self.on_trace_read(proto, {}, {})

        for idx, device_usage in enumerate(proto.device_usage):
            store = graphsignal._agent.metric_store()
            metric_tags = {'deployment': graphsignal._agent.deployment}
            if graphsignal._agent.hostname:
                metric_tags['hostname'] = graphsignal._agent.hostname
            metric_tags['device'] = idx

            if device_usage.gpu_utilization_percent > 0:
                store.set_gauge(
                    scope='system', name='gpu_utilization', tags=metric_tags, 
                    value=device_usage.gpu_utilization_percent, update_ts=now, unit='%')
            if device_usage.mxu_utilization_percent > 0:
                store.set_gauge(
                    scope='system', name='mxu_utilization', tags=metric_tags, 
                    value=device_usage.mxu_utilization_percent, update_ts=now, unit='%')
            if device_usage.mem_access_percent > 0:
                store.set_gauge(
                    scope='system', name='device_memory_access', tags=metric_tags, 
                    value=device_usage.mem_access_percent, update_ts=now, unit='%')
            if device_usage.mem_used > 0:
                store.set_gauge(
                    scope='system', name='device_memory_used', tags=metric_tags, 
                    value=device_usage.mem_used, update_ts=now, is_size=True)
            if device_usage.nvlink_throughput_data_tx_kibs > 0:
                store.set_gauge(
                    scope='system', name='nvlink_throughput_data_tx_kibs', tags=metric_tags, 
                    value=device_usage.nvlink_throughput_data_tx_kibs, update_ts=now, unit='KiB/s')
            if device_usage.nvlink_throughput_data_rx_kibs > 0:
                store.set_gauge(
                    scope='system', name='nvlink_throughput_data_rx_kibs', tags=metric_tags, 
                    value=device_usage.nvlink_throughput_data_rx_kibs, update_ts=now, unit='KiB/s')
            if device_usage.gpu_temp_c > 0:
                store.set_gauge(
                    scope='system', name='gpu_temp_c', tags=metric_tags, 
                    value=device_usage.gpu_temp_c, update_ts=now, unit='°C')
            if device_usage.power_usage_w > 0:
                store.set_gauge(
                    scope='system', name='power_usage_w', tags=metric_tags, 
                    value=device_usage.power_usage_w, update_ts=now, unit='W')

def _avg_sample_value(sample_value_type, samples):
    if not samples:
        return 0

    sample_values = []

    if sample_value_type == NVML_VALUE_TYPE_DOUBLE:
        sample_values = [sample.sampleValue.dVal for sample in samples]
    if sample_value_type == NVML_VALUE_TYPE_UNSIGNED_INT:
        sample_values = [sample.sampleValue.uiVal for sample in samples]
    if sample_value_type == NVML_VALUE_TYPE_UNSIGNED_LONG:
        sample_values = [sample.sampleValue.ulVal for sample in samples]
    if sample_value_type == NVML_VALUE_TYPE_UNSIGNED_LONG_LONG:
        sample_values = [sample.sampleValue.ullVal for sample in samples]

    if len(sample_values) > 0:
        return sum(sample_values) / len(sample_values)

    return 0


def _nvml_value(value_type, value):
    if value_type == NVML_VALUE_TYPE_DOUBLE:
        return value.dVal
    if value_type == NVML_VALUE_TYPE_UNSIGNED_INT:
        return value.uiVal
    if value_type == NVML_VALUE_TYPE_UNSIGNED_LONG:
        return value.ulVal
    if value_type == NVML_VALUE_TYPE_UNSIGNED_LONG_LONG:
        return value.ullVal


def _log_nvml_error(err):
    if (err.value == NVML_ERROR_NOT_SUPPORTED):
        logger.debug('NVML call not supported', exc_info=True)
    elif (err.value == NVML_ERROR_NOT_FOUND):
        logger.debug('NVML call not found', exc_info=True)
    else:
        logger.error('Error calling NVML', exc_info=True)


def _format_version(version):
    major = int(version / 1000)
    minor = int(version % 1000 / 10)
    return '{0}.{1}'.format(major, minor)



import unittest
import logging
import sys
import os
import json
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint
import tensorflow as tf

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader
from graphsignal.traces import DEFAULT_OPTIONS
from graphsignal.recorders.tensorflow_recorder import TensorFlowRecorder

logger = logging.getLogger('graphsignal')


class TensorFlowRecorderTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_record(self):
        os.environ["TF_CONFIG"] = json.dumps({
            "cluster": {
                "chief": ["host1:port"],
                "worker": ["host1:port", "host2:port"]
            },
            "task": {"type": "worker", "index": 1}
        })

        recorder = TensorFlowRecorder()
        recorder.setup()
        proto = signals_pb2.Trace()
        context = {}
        recorder.on_trace_start(proto, context, DEFAULT_OPTIONS)
        recorder.on_trace_stop(proto, context, DEFAULT_OPTIONS)
        recorder.on_trace_read(proto, context, DEFAULT_OPTIONS)

        self.assertEqual(proto.frameworks[0].name, 'TensorFlow')

        self.assertEqual(proto.frameworks[0].params[0].name, 'cluster_size')
        self.assertEqual(proto.frameworks[0].params[0].value, '3')
        self.assertEqual(proto.frameworks[0].params[1].name, 'task_index')
        self.assertEqual(proto.frameworks[0].params[1].value, '1')
        self.assertEqual(proto.frameworks[0].params[2].name, 'tf.test.is_built_with_gpu_support')
        self.assertEqual(proto.frameworks[0].params[2].value, str(tf.test.is_built_with_gpu_support()))



import logging
import functools

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.data.base_data_profiler import BaseDataProfiler, DataStats

logger = logging.getLogger('graphsignal')


class TorchTensorProfiler(BaseDataProfiler):
    def __init__(self):
        super().__init__()

    def is_instance(self, data):
        torch = self.check_module('torch')
        return torch is not None and torch.is_tensor(data)

    def compute_stats(self, data):
        torch = self.check_module('torch')
        counts = {}
        counts['element_count'] = data.nelement()
        counts['byte_count'] = counts['element_count'] * data.element_size()
        counts['nan_count'] = int(torch.count_nonzero(torch.isnan(data)))
        counts['inf_count'] = int(torch.count_nonzero(torch.isinf(data)))
        counts['zero_count'] = int(torch.count_nonzero(torch.eq(data, 0)))
        counts['negative_count'] = int(torch.count_nonzero(torch.lt(data, 0)))
        counts['positive_count'] = int(torch.count_nonzero(torch.gt(data, 0)))
        return DataStats(type_name='torch.Tensor', shape=list(data.size()), counts=counts)

    def encode_sample(self, data):
        return None
